@@ -127,7 +127,7 @@ static TYPELIB command_typelib=
 static struct my_option my_long_options[] =
 {
 #ifdef __NETWARE__
-  {"auto-close", OPT_AUTO_CLOSE, "Auto close the screen on exit for Netware.",
+  {"autoclose", OPT_AUTO_CLOSE, "Auto close the screen on exit for Netware.",
    0, 0, 0, GET_NO_ARG, NO_ARG, 0, 0, 0, 0, 0, 0},
 #endif
   {"count", 'c',
@@ -161,7 +161,7 @@ static struct my_option my_long_options[] =
    NO_ARG, 0, 0, 0, 0, 0, 0},
 #endif
   {"port", 'P', "Port number to use for connection.", (gptr*) &tcp_port,
-   (gptr*) &tcp_port, 0, GET_UINT, REQUIRED_ARG, MYSQL_PORT, 0, 0, 0, 0, 0},
+   (gptr*) &tcp_port, 0, GET_UINT, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   {"protocol", OPT_MYSQL_PROTOCOL, "The protocol of connection (tcp,socket,pipe,memory).",
     0, 0, 0, GET_STR,  REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   {"relative", 'r',
@@ -340,6 +340,8 @@ int main(int argc,char *argv[])
   if (opt_use_ssl)
     mysql_ssl_set(&mysql, opt_ssl_key, opt_ssl_cert, opt_ssl_ca,
 		  opt_ssl_capath, opt_ssl_cipher);
+  mysql_options(&mysql,MYSQL_OPT_SSL_VERIFY_SERVER_CERT,
+                (char*)&opt_ssl_verify_server_cert);
 #endif
   if (opt_protocol)
     mysql_options(&mysql,MYSQL_OPT_PROTOCOL,(char*)&opt_protocol);
@@ -431,6 +433,7 @@ static my_bool sql_connect(MYSQL *mysql, uint wait)
     if (mysql_real_connect(mysql,host,user,opt_password,NullS,tcp_port,
 			   unix_port, 0))
     {
+      mysql->reconnect= 1;
       if (info)
       {
 	fputs("\n",stderr);
@@ -568,6 +571,7 @@ static int execute_commands(MYSQL *mysql,int argc, char **argv)
 	return -1;
       }
       mysql_close(mysql);	/* Close connection to avoid error messages */
+      argc=1;                   /* force SHUTDOWN to be the last command    */
       if (got_pidfile)
       {
 	if (opt_verbose)
@@ -726,7 +730,7 @@ static int execute_commands(MYSQL *mysql,int argc, char **argv)
       void (*func) (MYSQL_RES*, MYSQL_ROW, uint);
 
       new_line = 1;
-      if (mysql_query(mysql, "show status") ||
+      if (mysql_query(mysql, "show /*!50002 GLOBAL */ status") ||
 	  !(res = mysql_store_result(mysql)))
       {
 	my_printf_error(0, "unable to show status; error: '%s'", MYF(ME_BELL),
@@ -1345,6 +1349,3 @@ static my_bool wait_pidfile(char *pidfile, time_t last_modified,
   }
   DBUG_RETURN(error);
 }
-#ifdef __GNUC__
-FIX_GCC_LINKING_PROBLEM
-#endif
