@@ -14,11 +14,29 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
 
+/* 
+
+   TODO: print the catalog (some USE catalog.db ????).
+
+   Standalone program to read a MySQL binary log (or relay log);
+   can read files produced by 3.23, 4.x, 5.0 servers. 
+
+   Can read binlogs from 3.23/4.x/5.0 and relay logs from 4.x/5.0.
+   Should be able to read any file of these categories, even with
+   --start-position.
+   An important fact: the Format_desc event of the log is at most the 3rd event
+   of the log; if it is the 3rd then there is this combination:
+   Format_desc_of_slave, Rotate_of_master, Format_desc_of_master.
+*/
+
 #define MYSQL_CLIENT
 #undef MYSQL_SERVER
 #include "client_priv.h"
 #include <my_time.h>
+/* That one is necessary for defines of OPTION_NO_FOREIGN_KEY_CHECKS etc */
+#include "mysql_priv.h" 
 #include "log_event.h"
+#include "sql_common.h"
 
 #define BIN_LOG_HEADER_SIZE	4
 #define PROBE_HEADER_LEN	(EVENT_LEN_OFFSET+4)
@@ -45,14 +63,16 @@ static const char *load_default_groups[]= { "mysqlbinlog","client",0 };
 void sql_print_error(const char *format, ...);
 
 static bool one_database=0, to_last_remote_log= 0, disable_log_bin= 0;
+static bool opt_hexdump= 0;
 static const char* database= 0;
 static my_bool force_opt= 0, short_form= 0, remote_opt= 0;
 static ulonglong offset = 0;
 static const char* host = 0;
-static int port = MYSQL_PORT;
+static int port= 0;
 static const char* sock= 0;
 static const char* user = 0;
 static char* pass = 0;
+static char *charset= 0;
 
 static ulonglong start_position, stop_position;
 #define start_position_mot ((my_off_t)start_position)
@@ -66,6 +86,14 @@ static MYSQL* mysql = NULL;
 static const char* dirname_for_local_load= 0;
 static bool stop_passed= 0;
 
+/*
+  check_header() will set the pointer below.
+  Why do we need here a pointer on an event instead of an event ?
+  This is because the event will be created (alloced) in read_log_event()
+  (which returns a pointer) in check_header().
+*/
+Format_description_log_event* description_event; 
+
 static int dump_local_log_entries(const char* logname);
 static int dump_remote_log_entries(const char* logname);
 static int dump_log_entries(const char* logname);
@@ -78,6 +106,22 @@ class Load_log_processor
 {
   char target_dir_name[FN_REFLEN];
   int target_dir_name_len;
+
+  /*
+    When we see first event corresponding to some LOAD DATA statement in
+    binlog, we create temporary file to store data to be loaded.
+    We add name of this file to file_names array using its file_id as index.
+    If we have Create_file event (i.e. we have binary log in pre-5.0.3
+    format) we also store save event object to be able which is needed to
+    emit LOAD DATA statement when we will meet Exec_load_data event.
+    If we have Begin_load_query event we simply store 0 in
+    File_name_record::event field.
+  */
+  struct File_name_record
+  {
+    char *fname;
+    Create_file_log_event *event;
+  };
   DYNAMIC_ARRAY file_names;
 
   /*
@@ -119,7 +163,7 @@ public:
 
   int init()
   {
-    return init_dynamic_array(&file_names,sizeof(Create_file_log_event*),
+    return init_dynamic_array(&file_names, sizeof(File_name_record),
 			      100,100 CALLER_INFO);
   }
 
@@ -136,33 +180,91 @@ public:
     }
   void destroy()
     {
-      Create_file_log_event **ptr= (Create_file_log_event**)file_names.buffer;
-      Create_file_log_event **end= ptr + file_names.elements;
+      File_name_record *ptr= (File_name_record *)file_names.buffer;
+      File_name_record *end= ptr + file_names.elements;
       for (; ptr<end; ptr++)
       {
-	if (*ptr)
+	if (ptr->fname)
 	{
-	  my_free((char*)(*ptr)->fname,MYF(MY_WME));
-	  delete *ptr;
-	  *ptr= 0;
+          my_free(ptr->fname, MYF(MY_WME));
+          delete ptr->event;
+          bzero((char *)ptr, sizeof(File_name_record));
 	}
       }
     }
+
+  /*
+    Obtain Create_file event for LOAD DATA statement by its file_id.
+
+    SYNOPSIS
+      grab_event()
+        file_id - file_id identifiying LOAD DATA statement
+
+    DESCRIPTION
+      Checks whenever we have already seen Create_file event for this file_id.
+      If yes then returns pointer to it and removes it from array describing
+      active temporary files. Since this moment caller is responsible for
+      freeing memory occupied by this event and associated file name.
+
+    RETURN VALUES
+      Pointer to Create_file event or 0 if there was no such event
+      with this file_id.
+  */
   Create_file_log_event *grab_event(uint file_id)
     {
+      File_name_record *ptr;
+      Create_file_log_event *res;
+
       if (file_id >= file_names.elements)
         return 0;
-      Create_file_log_event **ptr= 
-	(Create_file_log_event**)file_names.buffer + file_id;
-      Create_file_log_event *res= *ptr;
-      *ptr= 0;
+      ptr= dynamic_element(&file_names, file_id, File_name_record*);
+      if ((res= ptr->event))
+        bzero((char *)ptr, sizeof(File_name_record));
+      return res;
+    }
+
+  /*
+    Obtain file name of temporary file for LOAD DATA statement by its file_id.
+
+    SYNOPSIS
+      grab_fname()
+        file_id - file_id identifiying LOAD DATA statement
+
+    DESCRIPTION
+      Checks whenever we have already seen Begin_load_query event for this
+      file_id. If yes then returns file name of corresponding temporary file.
+      Removes record about this file from the array of active temporary files.
+      Since this moment caller is responsible for freeing memory occupied by
+      this name.
+
+    RETURN VALUES
+      String with name of temporary file or 0 if we have not seen Begin_load_query
+      event with this file_id.
+  */
+  char *grab_fname(uint file_id)
+    {
+      File_name_record *ptr;
+      char *res= 0;
+
+      if (file_id >= file_names.elements)
+        return 0;
+      ptr= dynamic_element(&file_names, file_id, File_name_record*);
+      if (!ptr->event)
+      {
+        res= ptr->fname;
+        bzero((char *)ptr, sizeof(File_name_record));
+      }
       return res;
     }
   int process(Create_file_log_event *ce);
+  int process(Begin_load_query_log_event *ce);
   int process(Append_block_log_event *ae);
   File prepare_new_file_for_old_format(Load_log_event *le, char *filename);
   int load_old_format_file(NET* net, const char *server_fname,
 			   uint server_fname_len, File file);
+  int process_first_event(const char *bname, uint blen, const char *block,
+                          uint block_len, uint file_id,
+                          Create_file_log_event *ce);
 };
 
 
@@ -240,22 +342,42 @@ int Load_log_processor::load_old_format_file(NET* net, const char*server_fname,
 }
 
 
-int Load_log_processor::process(Create_file_log_event *ce)
+/*
+  Process first event in the sequence of events representing LOAD DATA
+  statement.
+
+  SYNOPSIS
+    process_first_event()
+      bname     - base name for temporary file to be created
+      blen      - base name length
+      block     - first block of data to be loaded
+      block_len - first block length
+      file_id   - identifies LOAD DATA statement
+      ce        - pointer to Create_file event object if we are processing
+                  this type of event.
+
+  DESCRIPTION
+    Creates temporary file to be used in LOAD DATA and writes first block of
+    data to it. Registers its file name (and optional Create_file event)
+    in the array of active temporary files.
+
+  RETURN VALUES
+    0     - success
+    non-0 - error
+*/
+
+int Load_log_processor::process_first_event(const char *bname, uint blen,
+                                            const char *block, uint block_len,
+                                            uint file_id,
+                                            Create_file_log_event *ce)
 {
-  const char *bname= ce->fname+dirname_length(ce->fname);
-  uint blen= ce->fname_len - (bname-ce->fname);
   uint full_len= target_dir_name_len + blen + 9 + 9 + 1;
   int error= 0;
   char *fname, *ptr;
   File file;
-  DBUG_ENTER("Load_log_processor::process");
+  File_name_record rec;
+  DBUG_ENTER("Load_log_processor::process_first_event");
 
-  if (set_dynamic(&file_names,(gptr)&ce,ce->file_id))
-  {
-    sql_print_error("Could not construct local filename %s%s",
-		    target_dir_name,bname);
-    DBUG_RETURN(-1);
-  }
   if (!(fname= my_malloc(full_len,MYF(MY_WME))))
     DBUG_RETURN(-1);
 
@@ -263,7 +385,7 @@ int Load_log_processor::process(Create_file_log_event *ce)
   ptr= fname + target_dir_name_len;
   memcpy(ptr,bname,blen);
   ptr+= blen;
-  ptr+= my_sprintf(ptr,(ptr,"-%x",ce->file_id));
+  ptr+= my_sprintf(ptr, (ptr, "-%x", file_id));
 
   if ((file= create_unique_file(fname,ptr)) < 0)
   {
@@ -271,9 +393,21 @@ int Load_log_processor::process(Create_file_log_event *ce)
 		    target_dir_name,bname);
     DBUG_RETURN(-1);
   }
-  ce->set_fname_outside_temp_buf(fname,strlen(fname));
 
-  if (my_write(file,(byte*) ce->block,ce->block_len,MYF(MY_WME|MY_NABP)))
+  rec.fname= fname;
+  rec.event= ce;
+
+  if (set_dynamic(&file_names, (gptr)&rec, file_id))
+  {
+    sql_print_error("Could not construct local filename %s%s",
+		    target_dir_name, bname);
+    DBUG_RETURN(-1);
+  }
+
+  if (ce)
+    ce->set_fname_outside_temp_buf(fname, strlen(fname));
+
+  if (my_write(file, (byte*)block, block_len, MYF(MY_WME|MY_NABP)))
     error= -1;
   if (my_close(file, MYF(MY_WME)))
     error= -1;
@@ -281,19 +415,35 @@ int Load_log_processor::process(Create_file_log_event *ce)
 }
 
 
+int Load_log_processor::process(Create_file_log_event *ce)
+{
+  const char *bname= ce->fname + dirname_length(ce->fname);
+  uint blen= ce->fname_len - (bname-ce->fname);
+
+  return process_first_event(bname, blen, ce->block, ce->block_len,
+                             ce->file_id, ce);
+}
+
+
+int Load_log_processor::process(Begin_load_query_log_event *blqe)
+{
+  return process_first_event("SQL_LOAD_MB", 11, blqe->block, blqe->block_len,
+                             blqe->file_id, 0);
+}
+
+
 int Load_log_processor::process(Append_block_log_event *ae)
 {
   DBUG_ENTER("Load_log_processor::process");
-  Create_file_log_event* ce= ((ae->file_id < file_names.elements) ?
-			      *((Create_file_log_event**)file_names.buffer +
-				ae->file_id) :
-			      0);
+  const char* fname= ((ae->file_id < file_names.elements) ?
+                       dynamic_element(&file_names, ae->file_id,
+                                       File_name_record*)->fname : 0);
 
-  if (ce)
+  if (fname)
   {
     File file;
     int error= 0;
-    if (((file= my_open(ce->fname,
+    if (((file= my_open(fname,
 			O_APPEND|O_BINARY|O_WRONLY,MYF(MY_WME))) < 0))
       DBUG_RETURN(-1);
     if (my_write(file,(byte*)ae->block,ae->block_len,MYF(MY_WME|MY_NABP)))
@@ -305,8 +455,8 @@ int Load_log_processor::process(Append_block_log_event *ae)
 
   /*
     There is no Create_file event (a bad binlog or a big
-    --position). Assuming it's a big --position, we just do nothing and
-    print a warning.
+    --start-position). Assuming it's a big --start-position, we just do
+    nothing and print a warning.
   */
   fprintf(stderr,"Warning: ignoring Append_block as there is no \
 Create_file event for file_id: %u\n",ae->file_id);
@@ -316,30 +466,58 @@ Create_file event for file_id: %u\n",ae->file_id);
 
 Load_log_processor load_processor;
 
+
+static bool check_database(const char *log_dbname)
+{
+  return one_database &&
+         (log_dbname != NULL) &&
+         strcmp(log_dbname, database);
+}
+
+
 /*
+  Process an event
+
+  SYNOPSIS
+    process_event()
+
   RETURN
     0           ok and continue
     1           error and terminate
     -1          ok and terminate
-
+ 
   TODO
     This function returns 0 even in some error cases. This should be changed.
 */
-int process_event(char *last_db, Log_event *ev, my_off_t pos, int old_format)
+
+
+
+int process_event(PRINT_EVENT_INFO *print_event_info, Log_event *ev,
+                  my_off_t pos)
 {
   char ll_buff[21];
+  Log_event_type ev_type= ev->get_type_code();
   DBUG_ENTER("process_event");
+  print_event_info->short_form= short_form;
 
+  /*
+    Format events are not concerned by --offset and such, we always need to
+    read them to be able to process the wanted events.
+  */
   if ((rec_count >= offset) &&
-      ((my_time_t)(ev->when) >= start_datetime))
+      ((my_time_t)(ev->when) >= start_datetime) ||
+      (ev_type == FORMAT_DESCRIPTION_EVENT))
   {
-    /*
-      We have found an event after start_datetime, from now on print
-      everything (in case the binlog has timestamps increasing and decreasing,
-      we do this to avoid cutting the middle).
-    */
-    start_datetime= 0;
-    offset= 0; // print everything and protect against cycling rec_count
+    if (ev_type != FORMAT_DESCRIPTION_EVENT)
+    {
+      /*
+        We have found an event after start_datetime, from now on print
+        everything (in case the binlog has timestamps increasing and
+        decreasing, we do this to avoid cutting the middle).
+      */
+      start_datetime= 0;
+      offset= 0; // print everything and protect against cycling rec_count
+    }
     if (((my_time_t)(ev->when) >= stop_datetime)
         || (pos >= stop_position_mot))
     {
@@ -348,32 +526,29 @@ int process_event(char *last_db, Log_event *ev, my_off_t pos, int old_format)
     }
     if (!short_form)
       fprintf(result_file, "# at %s\n",llstr(pos,ll_buff));
-    
-    switch (ev->get_type_code()) {
+
+    if (!opt_hexdump)
+      print_event_info->hexdump_from= 0; /* Disabled */
+    else
+      print_event_info->hexdump_from= pos;
+
+    switch (ev_type) {
     case QUERY_EVENT:
-      if (one_database)
-      {
-	const char * log_dbname = ((Query_log_event*)ev)->db;
-	if ((log_dbname != NULL) && (strcmp(log_dbname, database)))
-	  goto end;
-      }
-      ev->print(result_file, short_form, last_db);
+      if (check_database(((Query_log_event*)ev)->db))
+        goto end;
+      ev->print(result_file, print_event_info);
       break;
     case CREATE_FILE_EVENT:
     {
       Create_file_log_event* ce= (Create_file_log_event*)ev;
-      if (one_database)
-      {
-	/*
-	  We test if this event has to be ignored. If yes, we don't save 
-	      this event; this will have the good side-effect of ignoring all 
-	      related Append_block and Exec_load.
-	      Note that Load event from 3.23 is not tested.
-	*/
-	const char * log_dbname = ce->db;            
-	if ((log_dbname != NULL) && (strcmp(log_dbname, database)))
-	  goto end;				// Next event
-      }
+      /*
+        We test if this event has to be ignored. If yes, we don't save
+        this event; this will have the good side-effect of ignoring all
+        related Append_block and Exec_load.
+        Note that Load event from 3.23 is not tested.
+      */
+      if (check_database(ce->db))
+        goto end;                // Next event
       /*
 	We print the event, but with a leading '#': this is just to inform 
 	the user of the original command; the command we want to execute 
@@ -381,8 +556,10 @@ int process_event(char *last_db, Log_event *ev, my_off_t pos, int old_format)
 	filename and use LOCAL), prepared in the 'case EXEC_LOAD_EVENT' 
 	below.
       */
-      ce->print(result_file, short_form, last_db, TRUE);
-      if (!old_format)
+      ce->print(result_file, print_event_info, TRUE);
+
+      // If this binlog is not 3.23 ; why this test??
+      if (description_event->binlog_version >= 3)
       {
 	if (load_processor.process(ce))
 	  break;				// Error
@@ -391,23 +568,23 @@ int process_event(char *last_db, Log_event *ev, my_off_t pos, int old_format)
       break;
     }
     case APPEND_BLOCK_EVENT:
-      ev->print(result_file, short_form, last_db);
+      ev->print(result_file, print_event_info);
       if (load_processor.process((Append_block_log_event*) ev))
 	break;					// Error
       break;
     case EXEC_LOAD_EVENT:
     {
-      ev->print(result_file, short_form, last_db);
+      ev->print(result_file, print_event_info);
       Execute_load_log_event *exv= (Execute_load_log_event*)ev;
       Create_file_log_event *ce= load_processor.grab_event(exv->file_id);
       /*
 	if ce is 0, it probably means that we have not seen the Create_file
-	event (a bad binlog, or most probably --position is after the
+	event (a bad binlog, or most probably --start-position is after the
 	Create_file event). Print a warning comment.
       */
       if (ce)
       {
-	ce->print(result_file, short_form, last_db, TRUE);
+	ce->print(result_file, print_event_info, TRUE);
 	my_free((char*)ce->fname,MYF(MY_WME));
 	delete ce;
       }
@@ -416,8 +593,47 @@ int process_event(char *last_db, Log_event *ev, my_off_t pos, int old_format)
 Create_file event for file_id: %u\n",exv->file_id);
       break;
     }
+    case FORMAT_DESCRIPTION_EVENT:
+      delete description_event;
+      description_event= (Format_description_log_event*) ev;
+      print_event_info->common_header_len= description_event->common_header_len;
+      ev->print(result_file, print_event_info);
+      /*
+        We don't want this event to be deleted now, so let's hide it (I
+        (Guilhem) should later see if this triggers a non-serious Valgrind
+        error). Not serious error, because we will free description_event
+        later.
+      */
+      ev= 0;
+      break;
+    case BEGIN_LOAD_QUERY_EVENT:
+      ev->print(result_file, print_event_info);
+      load_processor.process((Begin_load_query_log_event*) ev);
+      break;
+    case EXECUTE_LOAD_QUERY_EVENT:
+    {
+      Execute_load_query_log_event *exlq= (Execute_load_query_log_event*)ev;
+      char *fname= load_processor.grab_fname(exlq->file_id);
+
+      if (check_database(exlq->db))
+      {
+        if (fname)
+          my_free(fname, MYF(MY_WME));
+        goto end;
+      }
+
+      if (fname)
+      {
+	exlq->print(result_file, print_event_info, fname);
+	my_free(fname, MYF(MY_WME));
+      }
+      else
+	fprintf(stderr,"Warning: ignoring Execute_load_query as there is no \
+Begin_load_query event for file_id: %u\n", exlq->file_id);
+      break;
+    }
     default:
-      ev->print(result_file, short_form, last_db);
+      ev->print(result_file, print_event_info);
     }
   }
 
@@ -433,9 +649,18 @@ static struct my_option my_long_options[] =
 {
 
 #ifdef __NETWARE__
-  {"auto-close", OPT_AUTO_CLOSE, "Auto close the screen on exit for Netware.",
+  {"autoclose", OPT_AUTO_CLOSE, "Auto close the screen on exit for Netware.",
    0, 0, 0, GET_NO_ARG, NO_ARG, 0, 0, 0, 0, 0, 0},
 #endif
+  /*
+    mysqlbinlog needs charsets knowledge, to be able to convert a charset
+    number found in binlog to a charset name (to be able to print things
+    like this:
+    SET @`a`:=_cp850 0x4DFC6C6C6572 COLLATE `cp850_general_ci`;
+  */
+  {"character-sets-dir", OPT_CHARSETS_DIR,
+   "Directory where character sets are.", (gptr*) &charsets_dir,
+   (gptr*) &charsets_dir, 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
 #ifndef DBUG_OFF
   {"debug", '#', "Output debug log.", (gptr*) &default_dbug_option,
    (gptr*) &default_dbug_option, 0, GET_STR, OPT_ARG, 0, 0, 0, 0, 0, 0},
@@ -455,6 +680,9 @@ static struct my_option my_long_options[] =
    0, 0},
   {"help", '?', "Display this help and exit.",
    0, 0, 0, GET_NO_ARG, NO_ARG, 0, 0, 0, 0, 0, 0},
+  {"hexdump", 'H', "Augment output with hexadecimal and ASCII event dump.",
+   (gptr*) &opt_hexdump, (gptr*) &opt_hexdump, 0, GET_BOOL, NO_ARG,
+   0, 0, 0, 0, 0, 0},
   {"host", 'h', "Get the binlog from server.", (gptr*) &host, (gptr*) &host,
    0, GET_STR_ALLOC, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   {"offset", 'o', "Skip the first N entries.", (gptr*) &offset, (gptr*) &offset,
@@ -462,7 +690,7 @@ static struct my_option my_long_options[] =
   {"password", 'p', "Password to connect to remote server.",
    0, 0, 0, GET_STR, OPT_ARG, 0, 0, 0, 0, 0, 0},
   {"port", 'P', "Use port to connect to the remote server.",
-   (gptr*) &port, (gptr*) &port, 0, GET_INT, REQUIRED_ARG, MYSQL_PORT, 0, 0,
+   (gptr*) &port, (gptr*) &port, 0, GET_INT, REQUIRED_ARG, 0, 0, 0,
    0, 0, 0},
   {"position", 'j', "Deprecated. Use --start-position instead.",
    (gptr*) &start_position, (gptr*) &start_position, 0, GET_ULL,
@@ -481,6 +709,9 @@ static struct my_option my_long_options[] =
    "Used to reserve file descriptors for usage by this program",
    (gptr*) &open_files_limit, (gptr*) &open_files_limit, 0, GET_ULONG,
    REQUIRED_ARG, MY_NFILE, 8, OS_FILE_LIMIT, 0, 1, 0},
+  {"set-charset", OPT_SET_CHARSET,
+   "Add 'SET NAMES character_set' to the output.", (gptr*) &charset,
+   (gptr*) &charset, 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   {"short-form", 's', "Just show the queries, no extra info.",
    (gptr*) &short_form, (gptr*) &short_form, 0, GET_BOOL, NO_ARG, 0, 0, 0, 0,
    0, 0},
@@ -562,7 +793,8 @@ static void die(const char* fmt, ...)
   fprintf(stderr, "\n");
   va_end(args);
   cleanup();
-  my_end(0);
+  /* We cannot free DBUG, it is used in global destructors after exit(). */
+  my_end(MY_DONT_FREE_DBUG);
   exit(1);
 }
 
@@ -570,7 +802,7 @@ static void die(const char* fmt, ...)
 
 static void print_version()
 {
-  printf("%s Ver 3.0 for %s at %s\n", my_progname, SYSTEM_TYPE, MACHINE_TYPE);
+  printf("%s Ver 3.1 for %s at %s\n", my_progname, SYSTEM_TYPE, MACHINE_TYPE);
   NETWARE_SET_SCREEN_MODE(1);
 }
 
@@ -596,7 +828,7 @@ static my_time_t convert_str_to_timestamp(const char* str)
   int was_cut;
   MYSQL_TIME l_time;
   long dummy_my_timezone;
-  bool dummy_in_dst_time_gap;
+  my_bool dummy_in_dst_time_gap;
   /* We require a total specification (date AND time) */
   if (str_to_datetime(str, strlen(str), &l_time, 0, &was_cut) !=
       MYSQL_TIMESTAMP_DATETIME || was_cut)
@@ -711,6 +943,7 @@ static MYSQL* safe_connect()
     mysql_close(local_mysql);
     die("failed on connect: %s", errmsg);
   }
+  local_mysql->reconnect= 1;
   return local_mysql;
 }
 
@@ -722,12 +955,17 @@ static int dump_log_entries(const char* logname)
 }
 
 
-static int check_master_version(MYSQL* mysql)
+/*
+  This is not as smart as check_header() (used for local log); it will not work
+  for a binlog which mixes format. TODO: fix this.
+*/
+static int check_master_version(MYSQL* mysql,
+                                Format_description_log_event
+                                **description_event)
 {
   MYSQL_RES* res = 0;
   MYSQL_ROW row;
   const char* version;
-  int old_format = 0;
 
   if (mysql_query(mysql, "SELECT VERSION()") ||
       !(res = mysql_store_result(mysql)))
@@ -753,11 +991,18 @@ static int check_master_version(MYSQL* mysql)
 
   switch (*version) {
   case '3':
-    old_format = 1;
+    *description_event= new Format_description_log_event(1);
     break;
   case '4':
+    *description_event= new Format_description_log_event(3);
   case '5':
-    old_format = 0;
+    /*
+      The server is soon going to send us its Format_description log
+      event, unless it is a 5.0 server with 3.23 or 4.0 binlogs.
+      So we first assume that this is 4.0 (which is enough to read the
+      Format_desc event if one comes).
+    */
+    *description_event= new Format_description_log_event(3);
     break;
   default:
     sql_print_error("Master reported unrecognized MySQL version '%s'",
@@ -767,18 +1012,18 @@ static int check_master_version(MYSQL* mysql)
     return 1;
   }
   mysql_free_result(res);
-  return old_format;
+  return 0;
 }
 
 
 static int dump_remote_log_entries(const char* logname)
+
 {
   char buf[128];
-  char last_db[FN_REFLEN+1] = "";
+  PRINT_EVENT_INFO print_event_info;
   ulong len;
   uint logname_len;
   NET* net;
-  int old_format;
   int error= 0;
   my_off_t old_off= start_position_mot;
   char fname[FN_REFLEN+1];
@@ -791,7 +1036,18 @@ static int dump_remote_log_entries(const char* logname)
   */
   mysql= safe_connect();
   net= &mysql->net;
-  old_format = check_master_version(mysql);
+
+  if (check_master_version(mysql, &description_event))
+  {
+    fprintf(stderr, "Could not find server version");
+    DBUG_RETURN(1);
+  }
+  if (!description_event || !description_event->is_valid())
+  {
+    fprintf(stderr, "Invalid Format_description log event; \
+could be out of memory");
+    DBUG_RETURN(1);
+  }
 
   /*
     COM_BINLOG_DUMP accepts only 4 bytes for the position, so we are forced to
@@ -820,7 +1076,9 @@ static int dump_remote_log_entries(const char* logname)
   for (;;)
   {
     const char *error_msg;
-    len = net_safe_read(mysql);
+    Log_event *ev;
+
+    len= cli_safe_read(mysql);
     if (len == packet_error)
     {
       fprintf(stderr, "Got error reading packet from server: %s\n",
@@ -830,11 +1088,11 @@ static int dump_remote_log_entries(const char* logname)
     }
     if (len < 8 && net->read_pos[0] == 254)
       break; // end of data
-    DBUG_PRINT("info",( "len= %u, net->read_pos[5] = %d\n",
+    DBUG_PRINT("info",( "len: %lu, net->read_pos[5]: %d\n",
 			len, net->read_pos[5]));
-    Log_event *ev = Log_event::read_log_event((const char*) net->read_pos + 1 ,
-					      len - 1, &error_msg, old_format);
-    if (!ev)
+    if (!(ev= Log_event::read_log_event((const char*) net->read_pos + 1 ,
+                                        len - 1, &error_msg,
+                                        description_event)))
     {
       fprintf(stderr, "Could not construct log event object\n");
       error= 1;
@@ -842,25 +1100,27 @@ static int dump_remote_log_entries(const char* logname)
     }   
 
     Log_event_type type= ev->get_type_code();
-    if (!old_format || ( type != LOAD_EVENT && type != CREATE_FILE_EVENT))
+    if (description_event->binlog_version >= 3 ||
+        (type != LOAD_EVENT && type != CREATE_FILE_EVENT))
     {
-      if (ev->get_type_code() == ROTATE_EVENT)
+      /*
+        If this is a Rotate event, maybe it's the end of the requested binlog;
+        in this case we are done (stop transfer).
+        This is suitable for binlogs, not relay logs (but for now we don't read
+        relay logs remotely because the server is not able to do that). If one
+        day we read relay logs remotely, then we will have a problem with the
+        detection below: relay logs contain Rotate events which are about the
+        binlogs, so which would trigger the end-detection below.
+      */
+      if (type == ROTATE_EVENT)
       {
         Rotate_log_event *rev= (Rotate_log_event *)ev;
         /*
-          mysqld is sending us all its binlogs after the requested one, but we
-          don't want them.
           If this is a fake Rotate event, and not about our log, we can stop
           transfer. If this a real Rotate event (so it's not about our log,
           it's in our log describing the next log), we print it (because it's
           part of our log) and then we will stop when we receive the fake one
           soon.
-          This is suitable for binlogs, not relay logs (but for now we don't
-          read relay logs remotely because the server is not able to do
-          that). If one day we read relay logs remotely, then we will have a
-          problem with the detection below: relay logs contain Rotate events
-          which are about the binlogs, so which would trigger the end-detection
-          below.
         */
         if (rev->when == 0)
         {
@@ -883,7 +1143,7 @@ static int dump_remote_log_entries(const char* logname)
           len= 1; // fake Rotate, so don't increment old_off
         }
       }
-      if ((error= process_event(last_db,ev,old_off,old_format)))
+      if ((error= process_event(&print_event_info, ev, old_off)))
       {
 	error= ((error < 0) ? 0 : 1);
         goto err;
@@ -895,64 +1155,136 @@ static int dump_remote_log_entries(const char* logname)
       const char *old_fname= le->fname;
       uint old_len= le->fname_len;
       File file;
-
+      
       if ((file= load_processor.prepare_new_file_for_old_format(le,fname)) < 0)
       {
         error= 1;
         goto err;
       }
-
-      if ((error= process_event(last_db,ev,old_off,old_format)))
+      
+      if ((error= process_event(&print_event_info, ev, old_off)))
       {
-	my_close(file,MYF(MY_WME));
+ 	my_close(file,MYF(MY_WME));
 	error= ((error < 0) ? 0 : 1);
         goto err;
       }
-      if (load_processor.load_old_format_file(net,old_fname,old_len,file))
+      error= load_processor.load_old_format_file(net,old_fname,old_len,file);
+      my_close(file,MYF(MY_WME));
+      if (error)
       {
-	my_close(file,MYF(MY_WME));
         error= 1;
         goto err;
       }
-      my_close(file,MYF(MY_WME));
     }
-    
     /*
       Let's adjust offset for remote log as for local log to produce 
       similar text.
     */
     old_off+= len-1;
   }
+
 err:
   mysql_close(mysql);
   DBUG_RETURN(error);
 }
 
 
-static int check_header(IO_CACHE* file)
+static void check_header(IO_CACHE* file, 
+                        Format_description_log_event **description_event) 
 {
   byte header[BIN_LOG_HEADER_SIZE];
   byte buf[PROBE_HEADER_LEN];
-  int old_format=0;
-  DBUG_ENTER("check_header");
+  my_off_t tmp_pos, pos;
 
-  my_off_t pos = my_b_tell(file);
+  *description_event= new Format_description_log_event(3);
+  pos= my_b_tell(file);
   my_b_seek(file, (my_off_t)0);
   if (my_b_read(file, header, sizeof(header)))
     die("Failed reading header;  Probably an empty file");
   if (memcmp(header, BINLOG_MAGIC, sizeof(header)))
     die("File is not a binary log file");
-  if (!my_b_read(file, buf, sizeof(buf)))
+
+  /*
+    Imagine we are running with --start-position=1000. We still need
+    to know the binlog format's. So we still need to find, if there is
+    one, the Format_desc event, or to know if this is a 3.23
+    binlog. So we need to first read the first events of the log,
+    those around offset 4.  Even if we are reading a 3.23 binlog from
+    the start (no --start-position): we need to know the header length
+    (which is 13 in 3.23, 19 in 4.x) to be able to successfully print
+    the first event (Start_log_event_v3). So even in this case, we
+    need to "probe" the first bytes of the log *before* we do a real
+    read_log_event(). Because read_log_event() needs to know the
+    header's length to work fine.
+  */
+  for(;;)
   {
-    if (buf[4] == START_EVENT)
+    tmp_pos= my_b_tell(file); /* should be 4 the first time */
+    if (my_b_read(file, buf, sizeof(buf)))
     {
-      uint event_len;
-      event_len = uint4korr(buf + EVENT_LEN_OFFSET);
-      old_format = (event_len < (LOG_EVENT_HEADER_LEN + START_HEADER_LEN));
+      if (file->error)
+        die("\
+Could not read entry at offset %lu : Error in log format or read error",
+            tmp_pos); 
+      /*
+        Otherwise this is just EOF : this log currently contains 0-2
+        events.  Maybe it's going to be filled in the next
+        milliseconds; then we are going to have a problem if this a
+        3.23 log (imagine we are locally reading a 3.23 binlog which
+        is being written presently): we won't know it in
+        read_log_event() and will fail().  Similar problems could
+        happen with hot relay logs if --start-position is used (but a
+        --start-position which is posterior to the current size of the log).
+        These are rare problems anyway (reading a hot log + when we
+        read the first events there are not all there yet + when we
+        read a bit later there are more events + using a strange
+        --start-position).
+      */
+      break;
+    }
+    else
+    {
+      DBUG_PRINT("info",("buf[4]=%d", buf[4]));
+      /* always test for a Start_v3, even if no --start-position */
+      if (buf[4] == START_EVENT_V3)       /* This is 3.23 or 4.x */
+      {
+        if (uint4korr(buf + EVENT_LEN_OFFSET) < 
+            (LOG_EVENT_MINIMAL_HEADER_LEN + START_V3_HEADER_LEN))
+        {
+          /* This is 3.23 (format 1) */
+          delete *description_event;
+          *description_event= new Format_description_log_event(1);
+        }
+        break;
+      }
+      else if (tmp_pos >= start_position)
+        break;
+      else if (buf[4] == FORMAT_DESCRIPTION_EVENT) /* This is 5.0 */
+      {
+        my_b_seek(file, tmp_pos); /* seek back to event's start */
+        if (!(*description_event= (Format_description_log_event*) 
+              Log_event::read_log_event(file, *description_event)))
+          /* EOF can't be hit here normally, so it's a real error */
+          die("Could not read a Format_description_log_event event \
+at offset %lu ; this could be a log format error or read error",
+              tmp_pos); 
+        DBUG_PRINT("info",("Setting description_event"));
+      }
+      else if (buf[4] == ROTATE_EVENT)
+      {
+        Log_event *ev;
+        my_b_seek(file, tmp_pos); /* seek back to event's start */
+        if (!(ev= Log_event::read_log_event(file, *description_event)))
+          /* EOF can't be hit here normally, so it's a real error */
+          die("Could not read a Rotate_log_event event at offset %lu ;"
+              " this could be a log format error or read error", tmp_pos);
+        delete ev;
+      }
+      else
+        break;
     }
   }
   my_b_seek(file, pos);
-  DBUG_RETURN(old_format);
 }
 
 
@@ -960,13 +1292,10 @@ static int dump_local_log_entries(const char* logname)
 {
   File fd = -1;
   IO_CACHE cache,*file= &cache;
-  char last_db[FN_REFLEN+1];
+  PRINT_EVENT_INFO print_event_info;
   byte tmp_buff[BIN_LOG_HEADER_SIZE];
-  bool old_format = 0;
   int error= 0;
 
-  last_db[0]= 0;
-
   if (logname && logname[0] != '-')
   {
     if ((fd = my_open(logname, O_RDONLY | O_BINARY, MYF(MY_WME))) < 0)
@@ -975,65 +1304,67 @@ static int dump_local_log_entries(const char* logname)
 		      MYF(MY_WME | MY_NABP)))
     {
       my_close(fd, MYF(MY_WME));
-      exit(1);
+      return 1;
     }
-    old_format = check_header(file);
+    check_header(file, &description_event);
   }
-  else
+  else // reading from stdin;
   {
-    if (init_io_cache(file, fileno(result_file), 0, READ_CACHE, (my_off_t) 0,
+    if (init_io_cache(file, fileno(stdin), 0, READ_CACHE, (my_off_t) 0,
 		      0, MYF(MY_WME | MY_NABP | MY_DONT_CHECK_FILESIZE)))
       return 1;
-    old_format = check_header(file);
+    check_header(file, &description_event);
     if (start_position)
     {
-      /* skip 'start_position' characters from stdout */
+      /* skip 'start_position' characters from stdin */
       byte buff[IO_SIZE];
       my_off_t length,tmp;
       for (length= start_position_mot ; length > 0 ; length-=tmp)
       {
 	tmp=min(length,sizeof(buff));
 	if (my_b_read(file, buff, (uint) tmp))
-	{
-	  error= 1;
-	  goto end;
-	}
+        {
+          error= 1;
+          goto end;
+        }
       }
     }
-    file->pos_in_file= start_position_mot;
-    file->seek_not_done=0;
   }
 
-  if (!start_position)
+  if (!description_event || !description_event->is_valid())
+    die("Invalid Format_description log event; could be out of memory");
+
+  if (!start_position && my_b_read(file, tmp_buff, BIN_LOG_HEADER_SIZE))
   {
-    // Skip header
-    if (my_b_read(file, tmp_buff, BIN_LOG_HEADER_SIZE))
-    {
-      error= 1;
-      goto end;
-    }
+    error= 1;
+    goto end;
   }
-
   for (;;)
   {
     char llbuff[21];
     my_off_t old_off = my_b_tell(file);
 
-    Log_event* ev = Log_event::read_log_event(file, old_format);
+    Log_event* ev = Log_event::read_log_event(file, description_event);
     if (!ev)
     {
-      if (file->error)
+      /*
+        if binlog wasn't closed properly ("in use" flag is set) don't complain
+        about a corruption, but treat it as EOF and move to the next binlog.
+      */
+      if (description_event->flags & LOG_EVENT_BINLOG_IN_USE_F)
+        file->error= 0;
+      else if (file->error)
       {
-	fprintf(stderr,
-		"Could not read entry at offset %s:"
-		"Error in log format or read error\n",
-		llstr(old_off,llbuff));
-	error= 1;
+        fprintf(stderr,
+                "Could not read entry at offset %s:"
+                "Error in log format or read error\n",
+                llstr(old_off,llbuff));
+        error= 1;
       }
       // file->error == 0 means EOF, that's OK, we break in this case
       break;
     }
-    if ((error= process_event(last_db,ev,old_off,false)))
+    if ((error= process_event(&print_event_info, ev, old_off)))
     {
       if (error < 0)
         error= 0;
@@ -1045,6 +1376,7 @@ end:
   if (fd >= 0)
     my_close(fd, MYF(MY_WME));
   end_io_cache(file);
+  delete description_event;
   return error;
 }
 
@@ -1095,6 +1427,21 @@ int main(int argc, char** argv)
     fprintf(result_file,
             "/*!32316 SET @OLD_SQL_LOG_BIN=@@SQL_LOG_BIN, SQL_LOG_BIN=0*/;\n");
 
+  /*
+    In mysqlbinlog|mysql, don't want mysql to be disconnected after each
+    transaction (which would be the case with GLOBAL.COMPLETION_TYPE==2).
+  */
+  fprintf(result_file,
+          "/*!50003 SET @OLD_COMPLETION_TYPE=@@COMPLETION_TYPE,"
+          "COMPLETION_TYPE=0*/;\n");
+
+  if (charset)
+    fprintf(result_file,
+            "\n/*!40101 SET @OLD_CHARACTER_SET_CLIENT=@@CHARACTER_SET_CLIENT */;"
+            "\n/*!40101 SET @OLD_CHARACTER_SET_RESULTS=@@CHARACTER_SET_RESULTS */;"
+            "\n/*!40101 SET @OLD_COLLATION_CONNECTION=@@COLLATION_CONNECTION */;"  
+            "\n/*!40101 SET NAMES %s */;\n", charset);
+
   for (save_stop_position= stop_position, stop_position= ~(my_off_t)0 ;
        (--argc >= 0) && !stop_passed ; )
   {
@@ -1109,9 +1456,22 @@ int main(int argc, char** argv)
     start_position= BIN_LOG_HEADER_SIZE;
   }
 
+  /*
+    Issue a ROLLBACK in case the last printed binlog was crashed and had half
+    of transaction.
+  */
+  fprintf(result_file,
+          "# End of log file\nROLLBACK /* added by mysqlbinlog */;\n"
+          "/*!50003 SET COMPLETION_TYPE=@OLD_COMPLETION_TYPE*/;\n");
   if (disable_log_bin)
     fprintf(result_file, "/*!32316 SET SQL_LOG_BIN=@OLD_SQL_LOG_BIN*/;\n");
 
+  if (charset)
+    fprintf(result_file,
+            "/*!40101 SET CHARACTER_SET_CLIENT=@OLD_CHARACTER_SET_CLIENT */;\n"
+            "/*!40101 SET CHARACTER_SET_RESULTS=@OLD_CHARACTER_SET_RESULTS */;\n"
+            "/*!40101 SET COLLATION_CONNECTION=@OLD_COLLATION_CONNECTION */;\n");
+
   if (tmpdir.list)
     free_tmpdir(&tmpdir);
   if (result_file != stdout)
@@ -1119,7 +1479,8 @@ int main(int argc, char** argv)
   cleanup();
   free_defaults(defaults_argv);
   my_free_open_file_info();
-  my_end(0);
+  /* We cannot free DBUG, it is used in global destructors after exit(). */
+  my_end(MY_DONT_FREE_DBUG);
   exit(exit_value);
   DBUG_RETURN(exit_value);			// Keep compilers happy
 }
@@ -1129,10 +1490,14 @@ int main(int argc, char** argv)
   the server
 */
 
-#ifdef __WIN__
+#include "my_decimal.h"
+#include "decimal.c"
+
+#if defined(__WIN__) && !defined(CMAKE_BUILD)
+#include "my_decimal.cpp"
 #include "log_event.cpp"
 #else
+#include "my_decimal.cc"
 #include "log_event.cc"
 #endif
 
-FIX_GCC_LINKING_PROBLEM
