@@ -63,8 +63,9 @@ HP_INFO *heap_open(const char *name, int mode)
 #ifndef DBUG_OFF
   info->opt_flag= READ_CHECK_USED;		/* Check when changing */
 #endif
-  DBUG_PRINT("exit",("heap: %lx  reclength: %d  records_in_block: %d",
-		     info,share->reclength,share->block.records_in_block));
+  DBUG_PRINT("exit",("heap: 0x%lx  reclength: %d  records_in_block: %d",
+		     (long) info, share->reclength,
+                     share->block.records_in_block));
   DBUG_RETURN(info);
 }
 
@@ -82,7 +83,7 @@ HP_SHARE *hp_find_named_heap(const char *name)
     info= (HP_SHARE*) pos->data;
     if (!strcmp(name, info->name))
     {
-      DBUG_PRINT("exit", ("Old heap_database: %lx",info));
+      DBUG_PRINT("exit", ("Old heap_database: 0x%lx", (long) info));
       DBUG_RETURN(info);
     }
   }
