@@ -221,15 +221,19 @@ in the free list to the frames.
 5) When we have AWE enabled, we disable adaptive hash indexes.
 */
 
+/* Value in microseconds */
+static const int WAIT_FOR_READ	= 20000;
+
 buf_pool_t*	buf_pool = NULL; /* The buffer buf_pool of the database */
 
+#ifdef UNIV_DEBUG
 ulint		buf_dbg_counter	= 0; /* This is used to insert validation
 					operations in excution in the
 					debug version */
 ibool		buf_debug_prints = FALSE; /* If this is set TRUE,
 					the program prints info whenever
 					read-ahead or flush occurs */
-
+#endif /* UNIV_DEBUG */
 /************************************************************************
 Calculates a page checksum which is stored to the page when it is written
 to a file. Note that we must be careful to calculate the same value on
@@ -320,7 +324,10 @@ buf_page_is_corrupted(
 			fprintf(stderr,
 "  InnoDB: Error: page %lu log sequence number %lu %lu\n"
 "InnoDB: is in the future! Current system log sequence number %lu %lu.\n"
-"InnoDB: Your database may be corrupt.\n",
+"InnoDB: Your database may be corrupt or you may have copied the InnoDB\n"
+"InnoDB: tablespace but not the InnoDB log files. See\n"
+"InnoDB: http://dev.mysql.com/doc/refman/5.0/en/forcing-recovery.html\n"
+"InnoDB: for more information.\n",
 		        (ulong) mach_read_from_4(read_buf + FIL_PAGE_OFFSET),
 			(ulong) ut_dulint_get_high(
 				mach_read_from_8(read_buf + FIL_PAGE_LSN)),
@@ -331,33 +338,43 @@ buf_page_is_corrupted(
 		}
 	}
 #endif
-	old_checksum = buf_calc_page_old_checksum(read_buf);
+  
+  /* If we use checksums validation, make additional check before returning
+  TRUE to ensure that the checksum is not equal to BUF_NO_CHECKSUM_MAGIC which
+  might be stored by InnoDB with checksums disabled.
+     Otherwise, skip checksum calculation and return FALSE */
+  
+  if (srv_use_checksums) {
+    old_checksum = buf_calc_page_old_checksum(read_buf); 
 
-	old_checksum_field = mach_read_from_4(read_buf + UNIV_PAGE_SIZE
+    old_checksum_field = mach_read_from_4(read_buf + UNIV_PAGE_SIZE
 					- FIL_PAGE_END_LSN_OLD_CHKSUM);
 
-	/* There are 2 valid formulas for old_checksum_field:
-	1. Very old versions of InnoDB only stored 8 byte lsn to the start
-	and the end of the page.
-	2. Newer InnoDB versions store the old formula checksum there. */
+    /* There are 2 valid formulas for old_checksum_field:
+	  1. Very old versions of InnoDB only stored 8 byte lsn to the start
+	  and the end of the page.
+	  2. Newer InnoDB versions store the old formula checksum there. */
 	
-	if (old_checksum_field != mach_read_from_4(read_buf + FIL_PAGE_LSN)
-	    && old_checksum_field != old_checksum) {
+    if (old_checksum_field != mach_read_from_4(read_buf + FIL_PAGE_LSN)
+        && old_checksum_field != old_checksum
+        && old_checksum_field != BUF_NO_CHECKSUM_MAGIC) {
 
-		return(TRUE);
-	}
+      return(TRUE);
+    }
 
-	checksum = buf_calc_page_new_checksum(read_buf);
-	checksum_field = mach_read_from_4(read_buf + FIL_PAGE_SPACE_OR_CHKSUM);
+    checksum = buf_calc_page_new_checksum(read_buf);
+    checksum_field = mach_read_from_4(read_buf + FIL_PAGE_SPACE_OR_CHKSUM);
 
-	/* InnoDB versions < 4.0.14 and < 4.1.1 stored the space id
-	(always equal to 0), to FIL_PAGE_SPACE_SPACE_OR_CHKSUM */
+    /* InnoDB versions < 4.0.14 and < 4.1.1 stored the space id
+	  (always equal to 0), to FIL_PAGE_SPACE_SPACE_OR_CHKSUM */
 
-	if (checksum_field != 0 && checksum_field != checksum) {
-
-	        return(TRUE);
-	}
+    if (checksum_field != 0 && checksum_field != checksum
+        && checksum_field != BUF_NO_CHECKSUM_MAGIC) {
 
+      return(TRUE);
+    }
+  }
+  
 	return(FALSE);
 }
 
@@ -379,8 +396,10 @@ buf_page_print(
 	ut_print_buf(stderr, read_buf, UNIV_PAGE_SIZE);
 	fputs("InnoDB: End of page dump\n", stderr);
 
-	checksum = buf_calc_page_new_checksum(read_buf);
-	old_checksum = buf_calc_page_old_checksum(read_buf);
+	checksum = srv_use_checksums ?
+    buf_calc_page_new_checksum(read_buf) : BUF_NO_CHECKSUM_MAGIC;
+	old_checksum = srv_use_checksums ?
+    buf_calc_page_old_checksum(read_buf) : BUF_NO_CHECKSUM_MAGIC;
 
 	ut_print_timestamp(stderr);
 	fprintf(stderr, 
@@ -449,23 +468,32 @@ buf_block_init(
 	byte*		frame)	/* in: pointer to buffer frame, or NULL if in
 				the case of AWE there is no frame */
 {
+	block->magic_n = 0;
+
 	block->state = BUF_BLOCK_NOT_USED;
 	
 	block->frame = frame;
 
 	block->awe_info = NULL;
 
+	block->buf_fix_count = 0;
+	block->io_fix = 0;
+
 	block->modify_clock = ut_dulint_zero;
 	
 	block->file_page_was_freed = FALSE;
 
 	block->check_index_page_at_flush = FALSE;
+	block->index = NULL;
 
 	block->in_free_list = FALSE;
 	block->in_LRU_list = FALSE;
 
 	block->n_pointers = 0;
 
+	mutex_create(&block->mutex);
+	mutex_set_level(&block->mutex, SYNC_BUF_BLOCK);
+
 	rw_lock_create(&(block->lock));
 	ut_ad(rw_lock_validate(&(block->lock)));
 
@@ -547,7 +575,7 @@ buf_pool_init(
 		}
 		/*----------------------------------------*/
 	} else {
-		buf_pool->frame_mem = ut_malloc_low(
+		buf_pool->frame_mem = os_mem_alloc_large(
 					UNIV_PAGE_SIZE * (n_frames + 1),
 					TRUE, FALSE);
 	}
@@ -734,8 +762,15 @@ buf_awe_map_page_to_frame(
 	bck = UT_LIST_GET_LAST(buf_pool->awe_LRU_free_mapped);
 
 	while (bck) {	
-		if (bck->state == BUF_BLOCK_FILE_PAGE
-	    	    && (bck->buf_fix_count != 0 || bck->io_fix != 0)) {
+		ibool skip;
+
+		mutex_enter(&bck->mutex);
+
+		skip = (bck->state == BUF_BLOCK_FILE_PAGE
+			&& (bck->buf_fix_count != 0 || bck->io_fix != 0));
+
+		if (skip) {
+			mutex_exit(&bck->mutex);
 
 			/* We have to skip this */
 			bck = UT_LIST_GET_PREV(awe_LRU_free_mapped, bck);
@@ -768,6 +803,8 @@ buf_awe_map_page_to_frame(
 
 			buf_pool->n_pages_awe_remapped++;
 			
+			mutex_exit(&bck->mutex);
+
 			return;
 		}
 	}
@@ -806,13 +843,22 @@ buf_block_make_young(
 /*=================*/
 	buf_block_t*	block)	/* in: block to make younger */
 {
-	if (buf_pool->freed_page_clock >= block->freed_page_clock 
-				+ 1 + (buf_pool->curr_size / 1024)) {
+#ifdef UNIV_SYNC_DEBUG
+	ut_ad(!mutex_own(&(buf_pool->mutex)));
+#endif /* UNIV_SYNC_DEBUG */
 
+	/* Note that we read freed_page_clock's without holding any mutex:
+	this is allowed since the result is used only in heuristics */
+
+	if (buf_pool->freed_page_clock >= block->freed_page_clock 
+				+ 1 + (buf_pool->curr_size / 4)) {
+
+		mutex_enter(&buf_pool->mutex);
 		/* There has been freeing activity in the LRU list:
 		best to move to the head of the LRU list */
 
 		buf_LRU_make_block_young(block);
+		mutex_exit(&buf_pool->mutex);
 	}
 }
 
@@ -847,12 +893,16 @@ buf_block_free(
 /*===========*/
 	buf_block_t*	block)	/* in, own: block to be freed */
 {
-	ut_a(block->state != BUF_BLOCK_FILE_PAGE);
-
 	mutex_enter(&(buf_pool->mutex));
 
+	mutex_enter(&block->mutex);
+
+	ut_a(block->state != BUF_BLOCK_FILE_PAGE);
+
 	buf_LRU_block_free_non_file_page(block);
 
+	mutex_exit(&block->mutex);
+
 	mutex_exit(&(buf_pool->mutex));
 }
 
@@ -1071,9 +1121,8 @@ buf_page_get_gen(
 #endif
 	buf_pool->n_page_gets++;
 loop:
-	mutex_enter_fast(&(buf_pool->mutex));
-
 	block = NULL;
+	mutex_enter_fast(&(buf_pool->mutex));
 	
 	if (guess) {
 		block = buf_block_align(guess);
@@ -1111,6 +1160,8 @@ loop:
 		goto loop;
 	}
 
+	mutex_enter(&block->mutex);
+
 	ut_a(block->state == BUF_BLOCK_FILE_PAGE);
 
 	must_read = FALSE;
@@ -1120,9 +1171,9 @@ loop:
 		must_read = TRUE;
 
 		if (mode == BUF_GET_IF_IN_POOL) {
-
 			/* The page is only being read to buffer */
-			mutex_exit(&(buf_pool->mutex));
+			mutex_exit(&buf_pool->mutex);
+			mutex_exit(&block->mutex);
 
 			return(NULL);
 		}
@@ -1146,7 +1197,7 @@ loop:
 #else
 	buf_block_buf_fix_inc(block);
 #endif
-	buf_block_make_young(block);
+	mutex_exit(&buf_pool->mutex);
 
 	/* Check if this is the first access to the page */
 
@@ -1154,10 +1205,13 @@ loop:
 
 	block->accessed = TRUE;
 
+	mutex_exit(&block->mutex);
+
+	buf_block_make_young(block);
+
 #ifdef UNIV_DEBUG_FILE_ACCESSES
 	ut_a(block->file_page_was_freed == FALSE);
 #endif	
-	mutex_exit(&(buf_pool->mutex));
 
 #ifdef UNIV_DEBUG
 	buf_dbg_counter++;
@@ -1182,13 +1236,14 @@ loop:
 		}
 
 		if (!success) {
-			mutex_enter(&(buf_pool->mutex));
+			mutex_enter(&block->mutex);
 
 			block->buf_fix_count--;
+
+			mutex_exit(&block->mutex);
 #ifdef UNIV_SYNC_DEBUG
 			rw_lock_s_unlock(&(block->debug_latch));
 #endif			
-			mutex_exit(&(buf_pool->mutex));
 
 			return(NULL);
 		}
@@ -1199,18 +1254,16 @@ loop:
 			completes */
 
 		        for (;;) {
-			        mutex_enter(&(buf_pool->mutex));
+				mutex_enter(&block->mutex);
 
 		                if (block->io_fix == BUF_IO_READ) {
 
-				        mutex_exit(&(buf_pool->mutex));
+					mutex_exit(&block->mutex);
 				  
-				        /* Sleep 20 milliseconds */
-
-				        os_thread_sleep(20000);
+					os_thread_sleep(WAIT_FOR_READ);
 				} else {
 				  
-				       mutex_exit(&(buf_pool->mutex));
+					mutex_exit(&block->mutex);
 
 				       break;
 				}
@@ -1268,14 +1321,15 @@ buf_page_optimistic_get_func(
 
 	ut_ad(mtr && block);
 	ut_ad((rw_latch == RW_S_LATCH) || (rw_latch == RW_X_LATCH));
-	
-	mutex_enter(&(buf_pool->mutex));
 
 	/* If AWE is used, block may have a different frame now, e.g., NULL */
-	
-	if (block->state != BUF_BLOCK_FILE_PAGE || block->frame != guess) {
 
-		mutex_exit(&(buf_pool->mutex));
+	mutex_enter(&block->mutex);
+
+	if (UNIV_UNLIKELY(block->state != BUF_BLOCK_FILE_PAGE)
+	    || UNIV_UNLIKELY(block->frame != guess)) {
+
+		mutex_exit(&block->mutex);
 
 		return(FALSE);
 	}
@@ -1285,16 +1339,15 @@ buf_page_optimistic_get_func(
 #else
 	buf_block_buf_fix_inc(block);
 #endif
+	accessed = block->accessed;
+	block->accessed = TRUE;
+
+	mutex_exit(&block->mutex);
+
 	buf_block_make_young(block);
 
 	/* Check if this is the first access to the page */
 
-	accessed = block->accessed;
-
-	block->accessed = TRUE;
-
-	mutex_exit(&(buf_pool->mutex));
-
 	ut_ad(!ibuf_inside() || ibuf_page(block->space, block->offset));
 
 	if (rw_latch == RW_S_LATCH) {
@@ -1307,19 +1360,20 @@ buf_page_optimistic_get_func(
 		fix_type = MTR_MEMO_PAGE_X_FIX;
 	}
 
-	if (!success) {
-		mutex_enter(&(buf_pool->mutex));
+	if (UNIV_UNLIKELY(!success)) {
+		mutex_enter(&block->mutex);
 		
 		block->buf_fix_count--;
+
+		mutex_exit(&block->mutex);
+
 #ifdef UNIV_SYNC_DEBUG
 		rw_lock_s_unlock(&(block->debug_latch));
-#endif			
-		mutex_exit(&(buf_pool->mutex));
-
+#endif
 		return(FALSE);
 	}
 
-	if (!UT_DULINT_EQ(modify_clock, block->modify_clock)) {
+	if (UNIV_UNLIKELY(!UT_DULINT_EQ(modify_clock, block->modify_clock))) {
 #ifdef UNIV_SYNC_DEBUG
 		buf_page_dbg_add_level(block->frame, SYNC_NO_ORDER_CHECK);
 #endif /* UNIV_SYNC_DEBUG */
@@ -1329,14 +1383,15 @@ buf_page_optimistic_get_func(
 			rw_lock_x_unlock(&(block->lock));
 		}
 
-		mutex_enter(&(buf_pool->mutex));
+		mutex_enter(&block->mutex);
 		
 		block->buf_fix_count--;
+
+		mutex_exit(&block->mutex);
+
 #ifdef UNIV_SYNC_DEBUG
 		rw_lock_s_unlock(&(block->debug_latch));
-#endif			
-		mutex_exit(&(buf_pool->mutex));
-		
+#endif
 		return(FALSE);
 	}
 
@@ -1355,7 +1410,7 @@ buf_page_optimistic_get_func(
 #ifdef UNIV_DEBUG_FILE_ACCESSES
 	ut_a(block->file_page_was_freed == FALSE);
 #endif
-	if (!accessed) {
+	if (UNIV_UNLIKELY(!accessed)) {
 		/* In the case of a first access, try to apply linear
 		read-ahead */
 
@@ -1394,10 +1449,10 @@ buf_page_get_known_nowait(
 	ut_ad(mtr);
 	ut_ad((rw_latch == RW_S_LATCH) || (rw_latch == RW_X_LATCH));
 	
-	mutex_enter(&(buf_pool->mutex));
-
 	block = buf_block_align(guess);
 
+	mutex_enter(&block->mutex);
+
 	if (block->state == BUF_BLOCK_REMOVE_HASH) {
 	        /* Another thread is just freeing the block from the LRU list
 	        of the buffer pool: do not try to access this page; this
@@ -1406,7 +1461,7 @@ buf_page_get_known_nowait(
 		we have already removed it from the page address hash table
 		of the buffer pool. */
 
-	        mutex_exit(&(buf_pool->mutex));
+		mutex_exit(&block->mutex);
 
 		return(FALSE);
 	}
@@ -1418,12 +1473,12 @@ buf_page_get_known_nowait(
 #else
 	buf_block_buf_fix_inc(block);
 #endif
+	mutex_exit(&block->mutex);
+
 	if (mode == BUF_MAKE_YOUNG) {
 		buf_block_make_young(block);
 	}
 
-	mutex_exit(&(buf_pool->mutex));
-
 	ut_ad(!ibuf_inside() || (mode == BUF_KEEP_OLD));
 
 	if (rw_latch == RW_S_LATCH) {
@@ -1437,13 +1492,15 @@ buf_page_get_known_nowait(
 	}
 	
 	if (!success) {
-		mutex_enter(&(buf_pool->mutex));
+		mutex_enter(&block->mutex);
 		
 		block->buf_fix_count--;
+
+		mutex_exit(&block->mutex);
+
 #ifdef UNIV_SYNC_DEBUG
 		rw_lock_s_unlock(&(block->debug_latch));
 #endif		
-		mutex_exit(&(buf_pool->mutex));
 
 		return(FALSE);
 	}
@@ -1491,8 +1548,7 @@ buf_page_init_for_backup_restore(
 	block->offset 		= offset;
 
 	block->lock_hash_val	= 0;
-	block->lock_mutex	= NULL;
-	
+
 	block->freed_page_clock = 0;
 
 	block->newest_modification = ut_dulint_zero;
@@ -1524,6 +1580,7 @@ buf_page_init(
 {
 #ifdef UNIV_SYNC_DEBUG
 	ut_ad(mutex_own(&(buf_pool->mutex)));
+	ut_ad(mutex_own(&(block->mutex)));
 #endif /* UNIV_SYNC_DEBUG */
 	ut_a(block->state != BUF_BLOCK_FILE_PAGE);
 
@@ -1535,10 +1592,10 @@ buf_page_init(
 	block->offset 		= offset;
 
 	block->check_index_page_at_flush = FALSE;
+	block->index		= NULL;
 	
 	block->lock_hash_val	= lock_rec_hash(space, offset);
-	block->lock_mutex	= NULL;
-	
+
 	/* Insert into the hash table of file pages */
 
         if (buf_page_hash_get(space, offset)) {
@@ -1630,6 +1687,7 @@ buf_page_init_for_read(
 	ut_a(block);
 
 	mutex_enter(&(buf_pool->mutex));
+	mutex_enter(&block->mutex);
 
 	if (fil_tablespace_deleted_or_being_deleted_in_mem(space,
 							tablespace_version)) {
@@ -1642,7 +1700,9 @@ buf_page_init_for_read(
 		/* The page belongs to a space which has been deleted or is
 		being deleted, or the page is already in buf_pool, return */
 
+		mutex_exit(&block->mutex);
 		mutex_exit(&(buf_pool->mutex));
+
 		buf_block_free(block);
 
 		if (mode == BUF_READ_IBUF_PAGES_ONLY) {
@@ -1662,6 +1722,7 @@ buf_page_init_for_read(
 	buf_LRU_add_block(block, TRUE); 	/* TRUE == to old blocks */
 	
 	block->io_fix = BUF_IO_READ;
+
 	buf_pool->n_pend_reads++;
 	
 	/* We set a pass-type x-lock on the frame because then the same
@@ -1673,6 +1734,7 @@ buf_page_init_for_read(
 	
 	rw_lock_x_lock_gen(&(block->lock), BUF_IO_READ);
 	
+	mutex_exit(&block->mutex);
  	mutex_exit(&(buf_pool->mutex));
 
 	if (mode == BUF_READ_IBUF_PAGES_ONLY) {
@@ -1728,13 +1790,17 @@ buf_page_create(
 
 	/* If we get here, the page was not in buf_pool: init it there */
 
+#ifdef UNIV_DEBUG
 	if (buf_debug_prints) {
 		fprintf(stderr, "Creating space %lu page %lu to buffer\n",
 			(ulong) space, (ulong) offset);
 	}
+#endif /* UNIV_DEBUG */
 
 	block = free_block;
 	
+	mutex_enter(&block->mutex);
+
 	buf_page_init(space, offset, block);
 
 	/* The block must be put to the LRU list */
@@ -1745,13 +1811,15 @@ buf_page_create(
 #else
 	buf_block_buf_fix_inc(block);
 #endif
+	buf_pool->n_pages_created++;
+
+	mutex_exit(&(buf_pool->mutex));
+
 	mtr_memo_push(mtr, block, MTR_MEMO_BUF_FIX);
 
 	block->accessed = TRUE;
 	
-	buf_pool->n_pages_created++;
-
-	mutex_exit(&(buf_pool->mutex));
+	mutex_exit(&block->mutex);
 
 	/* Delete possible entries for the page from the insert buffer:
 	such can exist if the page belonged to an index which was dropped */
@@ -1762,6 +1830,15 @@ buf_page_create(
 	buf_flush_free_margin();
 
 	frame = block->frame;
+
+	/* Reset to zero the file flush lsn field in the page; if the first
+	page of an ibdata file is 'created' in this function into the buffer
+	pool then we lose the original contents of the file flush lsn stamp.
+	Then InnoDB could in a crash recovery print a big, false, corruption
+	warning if the stamp contains an lsn bigger than the ib_logfile lsn. */
+
+	memset(frame + FIL_PAGE_FILE_FLUSH_LSN, 0, 8);
+
 #ifdef UNIV_DEBUG
 	buf_dbg_counter++;
 
@@ -1791,6 +1868,12 @@ buf_page_io_complete(
 
 	ut_a(block->state == BUF_BLOCK_FILE_PAGE);
 
+	/* We do not need protect block->io_fix here by block->mutex to read
+	it because this is the only function where we can change the value
+	from BUF_IO_READ or BUF_IO_WRITE to some other value, and our code
+	ensures that this is the only thread that handles the i/o for this
+	block. */
+
 	io_type = block->io_fix;
 
 	if (io_type == BUF_IO_READ) {
@@ -1837,7 +1920,7 @@ buf_page_io_complete(
 		"InnoDB: the corrupt table. You can use CHECK\n"
 		"InnoDB: TABLE to scan your table for corruption.\n"
 		"InnoDB: See also "
-		"http://dev.mysql.com/doc/mysql/en/Forcing_recovery.html\n"
+"InnoDB: http://dev.mysql.com/doc/refman/5.0/en/forcing-recovery.html\n"
 		"InnoDB: about forcing recovery.\n", stderr);
 			  
 			if (srv_force_recovery < SRV_FORCE_IGNORE_CORRUPT) {
@@ -1859,11 +1942,12 @@ buf_page_io_complete(
 		}
 	}
 	
+	mutex_enter(&(buf_pool->mutex));
+	mutex_enter(&block->mutex);
+
 #ifdef UNIV_IBUF_DEBUG
 	ut_a(ibuf_count_get(block->space, block->offset) == 0);
 #endif
-	mutex_enter(&(buf_pool->mutex));
-	
 	/* Because this thread which does the unlocking is not the same that
 	did the locking, we use a pass value != 0 in unlock, which simply
 	removes the newest lock debug record, without checking the thread
@@ -1882,9 +1966,11 @@ buf_page_io_complete(
 
 		rw_lock_x_unlock_gen(&(block->lock), BUF_IO_READ);
 
+#ifdef UNIV_DEBUG
 		if (buf_debug_prints) {
 			fputs("Has read ", stderr);
 		}
+#endif /* UNIV_DEBUG */
 	} else {
 		ut_ad(io_type == BUF_IO_WRITE);
 
@@ -1897,17 +1983,22 @@ buf_page_io_complete(
 
 		buf_pool->n_pages_written++;
 
+#ifdef UNIV_DEBUG
 		if (buf_debug_prints) {
 			fputs("Has written ", stderr);
 		}
+#endif /* UNIV_DEBUG */
 	}
 	
+	mutex_exit(&block->mutex);
 	mutex_exit(&(buf_pool->mutex));
 
+#ifdef UNIV_DEBUG
 	if (buf_debug_prints) {
 		fprintf(stderr, "page space %lu page no %lu\n",
 			(ulong) block->space, (ulong) block->offset);
 	}
+#endif /* UNIV_DEBUG */
 }
 
 /*************************************************************************
@@ -1936,6 +2027,7 @@ buf_pool_invalidate(void)
 	mutex_exit(&(buf_pool->mutex));
 }
 
+#ifdef UNIV_DEBUG
 /*************************************************************************
 Validates the buffer buf_pool data structure. */
 
@@ -1961,6 +2053,8 @@ buf_validate(void)
 
 		block = buf_pool_get_nth_block(buf_pool, i);
 
+		mutex_enter(&block->mutex);
+
 		if (block->state == BUF_BLOCK_FILE_PAGE) {
 
 			ut_a(buf_page_hash_get(block->space,
@@ -2004,6 +2098,8 @@ buf_validate(void)
 		} else if (block->state == BUF_BLOCK_NOT_USED) {
 			n_free++;
 		}
+
+		mutex_exit(&block->mutex);
  	}
 
 	if (n_lru + n_free > buf_pool->curr_size) {
@@ -2080,10 +2176,6 @@ buf_print(void)
 	
 	n_found = 0;
 
-	for (i = 0 ; i < size; i++) {
-		counts[i] = 0;
-	}
-
 	for (i = 0; i < size; i++) {
 		frame = buf_pool_get_nth_block(buf_pool, i)->frame;
 
@@ -2135,6 +2227,38 @@ buf_print(void)
 
 	ut_a(buf_validate());
 }	
+#endif /* UNIV_DEBUG */
+
+/*************************************************************************
+Returns the number of latched pages in the buffer pool. */
+
+ulint
+buf_get_latched_pages_number(void)
+{
+        buf_block_t* block;
+        ulint i;
+        ulint fixed_pages_number = 0;
+
+        mutex_enter(&(buf_pool->mutex));
+
+        for (i = 0; i < buf_pool->curr_size; i++) {
+
+		block = buf_pool_get_nth_block(buf_pool, i);
+
+		if (block->magic_n == BUF_BLOCK_MAGIC_N) {
+			mutex_enter(&block->mutex);
+
+			if (block->buf_fix_count != 0 || block->io_fix != 0) {
+				fixed_pages_number++;
+			}
+
+			mutex_exit(&block->mutex);
+		}
+        }
+
+        mutex_exit(&(buf_pool->mutex));
+        return fixed_pages_number;
+}
 
 /*************************************************************************
 Returns the number of pending buf pool ios. */
@@ -2294,16 +2418,21 @@ buf_all_freed(void)
 
 		block = buf_pool_get_nth_block(buf_pool, i);
 
+		mutex_enter(&block->mutex);
+
 		if (block->state == BUF_BLOCK_FILE_PAGE) {
 
 			if (!buf_flush_ready_for_replace(block)) {
 
 				fprintf(stderr,
 					"Page %lu %lu still fixed or dirty\n",
-					(ulong) block->space, (ulong) block->offset);
+					(ulong) block->space,
+					(ulong) block->offset);
 			    	ut_error;
 			}
 		}
+
+		mutex_exit(&block->mutex);
  	}
 
 	mutex_exit(&(buf_pool->mutex));
