@@ -134,7 +134,12 @@ int opt_sum_query(TABLE_LIST *tables, List<Item> &all_fields,COND *conds)
     }
     else
     {
-      tl->table->file->info(HA_STATUS_VARIABLE | HA_STATUS_NO_LOCK);
+      error= tl->table->file->info(HA_STATUS_VARIABLE | HA_STATUS_NO_LOCK);
+      if(error)
+      {
+        tl->table->file->print_error(error, MYF(0));
+        return error;
+      }
       count*= tl->table->file->records;
     }
   }
@@ -191,7 +196,7 @@ int opt_sum_query(TABLE_LIST *tables, List<Item> &all_fields,COND *conds)
             Type of range for the key part for this field will be
             returned in range_fl.
           */
-          if ((outer_tables & table->map) ||
+          if (table->file->inited || (outer_tables & table->map) ||
               !find_key_for_maxmin(0, &ref, item_field->field, conds,
                                    &range_fl, &prefix_len))
           {
@@ -278,7 +283,7 @@ int opt_sum_query(TABLE_LIST *tables, List<Item> &all_fields,COND *conds)
             Type of range for the key part for this field will be
             returned in range_fl.
           */
-          if ((outer_tables & table->map) ||
+          if (table->file->inited || (outer_tables & table->map) ||
 	          !find_key_for_maxmin(1, &ref, item_field->field, conds,
 				                   &range_fl, &prefix_len))
           {
@@ -608,7 +613,8 @@ static bool matching_cond(bool max_fl, TABLE_REF *ref, KEY *keyinfo,
     }
     else
     {
-      store_val_in_field(part->field, args[between && max_fl ? 2 : 1]);
+      store_val_in_field(part->field, args[between && max_fl ? 2 : 1],
+                         CHECK_FIELD_IGNORE);
       if (part->null_bit) 
         *key_ptr++= (byte) test(part->field->is_null());
       part->field->get_key_image((char*) key_ptr, part->length, Field::itRAW);
@@ -663,6 +669,8 @@ static bool matching_cond(bool max_fl, TABLE_REF *ref, KEY *keyinfo,
         field BETWEEN const1 AND const2
      3. all references to the columns from the same table as column field
         occur only in conjucts mentioned above.
+     4. each of k first components the index is not partial, i.e. is not
+        defined on a fixed length proper prefix of the field.
 
      If such an index exists the function through the ref parameter
      returns the key value to find max/min for the field using the index,
@@ -672,8 +680,8 @@ static bool matching_cond(bool max_fl, TABLE_REF *ref, KEY *keyinfo,
       of the whole search key)
 
   NOTE
-   This function may set table->key_read to 1, which must be reset after
-   index is used! (This can only happen when function returns 1)
+    This function may set table->key_read to 1, which must be reset after
+    index is used! (This can only happen when function returns 1)
 
   RETURN
     0   Index can not be used to optimize MIN(field)/MAX(field)
@@ -698,6 +706,12 @@ static bool find_key_for_maxmin(bool max_fl, TABLE_REF *ref,
   {
     KEY_PART_INFO *part,*part_end;
     key_part_map key_part_to_use= 0;
+    /*
+      Perform a check if index is not disabled by ALTER TABLE
+      or IGNORE INDEX.
+    */
+    if (!table->keys_in_use_for_query.is_set(idx))
+      continue;
     uint jdx= 0;
     *prefix_len= 0;
     for (part= keyinfo->key_part, part_end= part+keyinfo->key_parts ;
@@ -707,6 +721,12 @@ static bool find_key_for_maxmin(bool max_fl, TABLE_REF *ref,
       if (!(table->file->index_flags(idx, jdx, 0) & HA_READ_ORDER))
         return 0;
 
+      /* Check whether the index component is partial */
+      Field *part_field= table->field[part->fieldnr-1];
+      if ((part_field->flags & BLOB_FLAG) ||
+          part->length < part_field->key_length())
+        break;
+
       if (field->eq(part->field))
       {
         ref->key= idx;
