@@ -179,9 +179,9 @@ THD::THD()
    lock_id(&main_lock_id),
    user_time(0), in_sub_stmt(0), global_read_lock(0), is_fatal_error(0),
    rand_used(0), time_zone_used(0),
-   last_insert_id_used(0), insert_id_used(0), clear_next_insert_id(0),
-   in_lock_tables(0), bootstrap(0), derived_tables_processing(FALSE),
-   spcont(NULL)
+   last_insert_id_used(0), last_insert_id_used_bin_log(0), insert_id_used(0),
+   clear_next_insert_id(0), in_lock_tables(0), bootstrap(0),
+   derived_tables_processing(FALSE), spcont(NULL)
 {
   stmt_arena= this;
   thread_stack= 0;
@@ -205,6 +205,7 @@ THD::THD()
   // Must be reset to handle error with THD's created for init of mysqld
   lex->current_select= 0;
   start_time=(time_t) 0;
+  time_after_lock=(time_t) 0;
   current_linfo =  0;
   slave_thread = 0;
   variables.pseudo_thread_id= 0;
@@ -224,7 +225,7 @@ THD::THD()
 #endif
   client_capabilities= 0;                       // minimalistic client
   net.last_error[0]=0;                          // If error on boot
-  net.query_cache_query=0;                      // If error on boot
+  query_cache_init_query(&net);                 // If error on boot
   ull=0;
   system_thread= cleanup_done= abort_on_warning= no_warnings_for_error= 0;
   peer_port= 0;					// For SHOW PROCESSLIST
@@ -270,6 +271,7 @@ THD::THD()
   tablespace_op=FALSE;
   ulong tmp=sql_rnd_with_mutex();
   randominit(&rand, tmp + (ulong) &rand, tmp + (ulong) ::query_id);
+  substitute_null_with_insert_id = FALSE;
   thr_lock_info_init(&lock_info); /* safety: will be reset after start */
   thr_lock_owner_init(&main_lock_id, &lock_info);
 }
@@ -307,6 +309,7 @@ void THD::init(void)
   total_warn_count= 0;
   update_charset();
   bzero((char *) &status_var, sizeof(status_var));
+  variables.lc_time_names = &my_locale_en_US;
 }
 
 
@@ -461,14 +464,13 @@ THD::~THD()
 
 void add_to_status(STATUS_VAR *to_var, STATUS_VAR *from_var)
 {
-  ulong *end= (ulong*) ((byte*) to_var + offsetof(STATUS_VAR,
-						  last_system_status_var) +
+  ulong *end= (ulong*) ((byte*) to_var +
+                        offsetof(STATUS_VAR, last_system_status_var) +
 			sizeof(ulong));
   ulong *to= (ulong*) to_var, *from= (ulong*) from_var;
 
   while (to != end)
     *(to++)+= *(from++);
-  /* it doesn't make sense to add last_query_cost values */
 }
 
 
@@ -551,10 +553,25 @@ bool THD::store_globals()
 }
 
 
-/* Cleanup after a query */
+/*
+  Cleanup after query.
+
+  SYNOPSIS
+    THD::cleanup_after_query()
+
+  DESCRIPTION
+    This function is used to reset thread data to its default state.
+
+  NOTE
+    This function is not suitable for setting thread data to some
+    non-default values, as there is only one replication thread, so
+    different master threads may overwrite data of each other on
+    slave.
+*/
 
 void THD::cleanup_after_query()
 {
+  last_insert_id_used= FALSE;
   if (clear_next_insert_id)
   {
     clear_next_insert_id= 0;
@@ -566,6 +583,7 @@ void THD::cleanup_after_query()
   where= THD::DEFAULT_WHERE;
 }
 
+
 /*
   Convert a string to another character set
 
@@ -694,7 +712,7 @@ void THD::add_changed_table(const char *key, long key_length)
     {
       list_include(prev_changed, curr, changed_table_dup(key, key_length));
       DBUG_PRINT("info", 
-		 ("key_length %u %u", key_length, (*prev_changed)->key_length));
+		 ("key_length %ld %u", key_length, (*prev_changed)->key_length));
       DBUG_VOID_RETURN;
     }
     else if (cmp == 0)
@@ -704,7 +722,7 @@ void THD::add_changed_table(const char *key, long key_length)
       {
 	list_include(prev_changed, curr, changed_table_dup(key, key_length));
 	DBUG_PRINT("info", 
-		   ("key_length %u %u", key_length,
+		   ("key_length %ld %u", key_length,
 		    (*prev_changed)->key_length));
 	DBUG_VOID_RETURN;
       }
@@ -716,7 +734,7 @@ void THD::add_changed_table(const char *key, long key_length)
     }
   }
   *prev_changed = changed_table_dup(key, key_length);
-  DBUG_PRINT("info", ("key_length %u %u", key_length,
+  DBUG_PRINT("info", ("key_length %ld %u", key_length,
 		      (*prev_changed)->key_length));
   DBUG_VOID_RETURN;
 }
@@ -931,7 +949,7 @@ bool select_send::send_data(List<Item> &items)
   Protocol *protocol= thd->protocol;
   char buff[MAX_FIELD_WIDTH];
   String buffer(buff, sizeof(buff), &my_charset_bin);
-  DBUG_ENTER("send_data");
+  DBUG_ENTER("select_send::send_data");
 
   protocol->prepare_for_resend();
   Item *item;
@@ -1141,7 +1159,7 @@ select_export::prepare(List<Item> &list, SELECT_LEX_UNIT *u)
 bool select_export::send_data(List<Item> &items)
 {
 
-  DBUG_ENTER("send_data");
+  DBUG_ENTER("select_export::send_data");
   char buff[MAX_FIELD_WIDTH],null_buff[2],space[MAX_FIELD_WIDTH];
   bool space_inited=0;
   String tmp(buff,sizeof(buff),&my_charset_bin),*res;
@@ -1298,7 +1316,7 @@ bool select_dump::send_data(List<Item> &items)
   String tmp(buff,sizeof(buff),&my_charset_bin),*res;
   tmp.length(0);
   Item *item;
-  DBUG_ENTER("send_data");
+  DBUG_ENTER("select_dump::send_data");
 
   if (unit->offset_limit_cnt)
   {						// using limit offset,count
@@ -1492,53 +1510,21 @@ bool select_exists_subselect::send_data(List<Item> &items)
 
 int select_dumpvar::prepare(List<Item> &list, SELECT_LEX_UNIT *u)
 {
-  List_iterator_fast<Item> li(list);
-  List_iterator_fast<my_var> gl(var_list);
-  Item *item;
-
-  local_vars.empty();				// Clear list if SP
   unit= u;
-  row_count= 0;
-
+  
   if (var_list.elements != list.elements)
   {
     my_message(ER_WRONG_NUMBER_OF_COLUMNS_IN_SELECT,
                ER(ER_WRONG_NUMBER_OF_COLUMNS_IN_SELECT), MYF(0));
     return 1;
-  }
-  while ((item=li++))
-  {
-    my_var *mv= gl++;
-    if (mv->local)
-    {
-      Item_splocal *var= new Item_splocal(mv->s, mv->offset, mv->type);
-      (void)local_vars.push_back(var);
-#ifndef DBUG_OFF
-      var->m_sp= mv->sp;
-#endif
-    }
-    else
-    {
-      Item_func_set_user_var *var= new Item_func_set_user_var(mv->s, item);
-      /*
-        Item_func_set_user_var can't substitute something else on its place =>
-        0 can be passed as last argument (reference on item)
-        Item_func_set_user_var can't be fixed after creation, so we do not
-        check var->fixed
-      */
-      var->fix_fields(thd, 0);
-      var->fix_length_and_dec();
-      vars.push_back(var);
-    }
-  }
+  }               
   return 0;
 }
 
 
 void select_dumpvar::cleanup()
 {
-  vars.empty();
-  row_count=0;
+  row_count= 0;
 }
 
 
@@ -1847,13 +1833,10 @@ Statement_map::~Statement_map()
 
 bool select_dumpvar::send_data(List<Item> &items)
 {
-  List_iterator_fast<Item_func_set_user_var> li(vars);
-  List_iterator_fast<Item_splocal> var_li(local_vars);
-  List_iterator_fast<my_var> my_li(var_list);
+  List_iterator_fast<my_var> var_li(var_list);
   List_iterator<Item> it(items);
-  Item_func_set_user_var *xx;
-  Item_splocal *yy;
-  my_var *zz;
+  Item *item;
+  my_var *mv;
   DBUG_ENTER("select_dumpvar::send_data");
 
   if (unit->offset_limit_cnt)
@@ -1866,24 +1849,19 @@ bool select_dumpvar::send_data(List<Item> &items)
     my_message(ER_TOO_MANY_ROWS, ER(ER_TOO_MANY_ROWS), MYF(0));
     DBUG_RETURN(1);
   }
-  while ((zz=my_li++) && (it++))
+  while ((mv= var_li++) && (item= it++))
   {
-    if (zz->local)
+    if (mv->local)
     {
-      if ((yy=var_li++)) 
-      {
-	if (thd->spcont->set_variable(current_thd, yy->get_var_idx(),
-                                      it.ref()))
-	  DBUG_RETURN(1);
-      }
+      if (thd->spcont->set_variable(thd, mv->offset, &item))
+	    DBUG_RETURN(1);
     }
     else
     {
-      if ((xx=li++))
-      {
-        xx->check();
-	xx->update();
-      }
+      Item_func_set_user_var *suv= new Item_func_set_user_var(mv->s, item);
+      suv->fix_fields(thd, 0);
+      suv->check(0);
+      suv->update();
     }
   }
   DBUG_RETURN(0);
@@ -1947,6 +1925,7 @@ void Security_context::init()
 {
   host= user= priv_user= ip= 0;
   host_or_ip= "connecting host";
+  priv_host[0]= '\0';
 #ifndef NO_EMBEDDED_ACCESS_CHECKS
   db_access= NO_ACCESS;
 #endif
