@@ -23,15 +23,18 @@
 
 #include "mysql_priv.h"
 #include "sql_select.h"
+#include "sql_cursor.h"
 
-int mysql_union(THD *thd, LEX *lex, select_result *result,
-		SELECT_LEX_UNIT *unit)
+bool mysql_union(THD *thd, LEX *lex, select_result *result,
+                 SELECT_LEX_UNIT *unit, ulong setup_tables_done_option)
 {
   DBUG_ENTER("mysql_union");
-  int res= 0;
-  if (!(res= unit->prepare(thd, result, SELECT_NO_UNLOCK, "")))
+  bool res;
+  if (!(res= unit->prepare(thd, result, SELECT_NO_UNLOCK |
+                           setup_tables_done_option)))
     res= unit->exec();
-  res|= unit->cleanup();
+  if (res || !thd->cursor || !thd->cursor->is_open())
+    res|= unit->cleanup();
   DBUG_RETURN(res);
 }
 
@@ -40,22 +43,6 @@ int mysql_union(THD *thd, LEX *lex, select_result *result,
 ** store records in temporary table for UNION
 ***************************************************************************/
 
-select_union::select_union(TABLE *table_par)
-  :table(table_par)
-{
-  bzero((char*) &info,sizeof(info));
-  /*
-    We can always use IGNORE because the temporary table will only
-    contain a unique key if we are using not using UNION ALL
-  */
-  info.ignore= 1;
-}
-
-select_union::~select_union()
-{
-}
-
-
 int select_union::prepare(List<Item> &list, SELECT_LEX_UNIT *u)
 {
   unit= u;
@@ -65,22 +52,21 @@ int select_union::prepare(List<Item> &list, SELECT_LEX_UNIT *u)
 
 bool select_union::send_data(List<Item> &values)
 {
+  int error= 0;
   if (unit->offset_limit_cnt)
   {						// using limit offset,count
     unit->offset_limit_cnt--;
     return 0;
   }
-  fill_record(table->field, values, 1);
-  if (thd->net.report_error || write_record(table,&info))
+  fill_record(thd, table->field, values, 1);
+  if (thd->net.report_error)
+    return 1;
+
+  if ((error= table->file->write_row(table->record[0])))
   {
-    if (thd->net.last_errno == ER_RECORD_FILE_FULL)
-    {
-      thd->clear_error(); // do not report user about table overflow
-      if (create_myisam_from_heap(thd, table, &tmp_table_param,
-				  info.last_errno, 1))
-	return 1;
-    }
-    else
+    /* create_myisam_from_heap will generate error if needed */
+    if (error != HA_ERR_FOUND_DUPP_KEY && error != HA_ERR_FOUND_DUPP_UNIQUE &&
+        create_myisam_from_heap(thd, table, &tmp_table_param, error, 1))
       return 1;
   }
   return 0;
@@ -98,13 +84,51 @@ bool select_union::flush()
   int error;
   if ((error=table->file->extra(HA_EXTRA_NO_CACHE)))
   {
-    table->file->print_error(error,MYF(0));
-    ::send_error(thd);
+    table->file->print_error(error, MYF(0));
     return 1;
   }
   return 0;
 }
 
+/*
+  Create a temporary table to store the result of select_union.
+
+  SYNOPSIS
+    select_union::create_result_table()
+      thd                thread handle
+      column_types       a list of items used to define columns of the
+                         temporary table
+      is_union_distinct  if set, the temporary table will eliminate
+                         duplicates on insert
+      options            create options
+
+  DESCRIPTION
+    Create a temporary table that is used to store the result of a UNION,
+    derived table, or a materialized cursor.
+
+  RETURN VALUE
+    0                    The table has been created successfully.
+    1                    create_tmp_table failed.
+*/
+
+bool
+select_union::create_result_table(THD *thd, List<Item> *column_types,
+                                  bool is_union_distinct, ulonglong options,
+                                  const char *alias)
+{
+  DBUG_ASSERT(table == 0);
+  tmp_table_param.init();
+  tmp_table_param.field_count= column_types->elements;
+
+  if (! (table= create_tmp_table(thd, &tmp_table_param, *column_types,
+                                 (ORDER*) 0, is_union_distinct, 1,
+                                 options, HA_POS_ERROR, (char*) alias)))
+    return TRUE;
+  table->file->extra(HA_EXTRA_WRITE_CACHE);
+  table->file->extra(HA_EXTRA_IGNORE_DUP_KEY);
+  return FALSE;
+}
+
 
 /*
   initialization procedures before fake_select_lex preparation()
@@ -117,36 +141,28 @@ bool select_union::flush()
     options of SELECT
 */
 
-ulong
+void
 st_select_lex_unit::init_prepare_fake_select_lex(THD *thd) 
 {
-  ulong options_tmp= thd->options | fake_select_lex->options;
   thd->lex->current_select= fake_select_lex;
-  offset_limit_cnt= global_parameters->offset_limit;
-  select_limit_cnt= global_parameters->select_limit +
-    global_parameters->offset_limit;
-
-  if (select_limit_cnt < global_parameters->select_limit)
-    select_limit_cnt= HA_POS_ERROR;		// no limit
-  if (select_limit_cnt == HA_POS_ERROR)
-    options_tmp&= ~OPTION_FOUND_ROWS;
-  else if (found_rows_for_union && !thd->lex->describe)
-    options_tmp|= OPTION_FOUND_ROWS;
-  fake_select_lex->ftfunc_list_alloc.empty();
-  fake_select_lex->ftfunc_list= &fake_select_lex->ftfunc_list_alloc;
   fake_select_lex->table_list.link_in_list((byte *)&result_table_list,
 					   (byte **)
-					   &result_table_list.next);
-  return options_tmp;
+					   &result_table_list.next_local);
+  for (ORDER *order= (ORDER *)global_parameters->order_list.first;
+       order;
+       order=order->next)
+  {
+    (*order->item)->walk(&Item::change_context_processor,
+                         (byte *) &fake_select_lex->context);
+  }
 }
 
 
-int st_select_lex_unit::prepare(THD *thd_arg, select_result *sel_result,
-				ulong additional_options,
-                                const char *tmp_table_alias)
+bool st_select_lex_unit::prepare(THD *thd_arg, select_result *sel_result,
+                                 ulong additional_options)
 {
   SELECT_LEX *lex_select_save= thd_arg->lex->current_select;
-  SELECT_LEX *sl, *first_select;
+  SELECT_LEX *sl, *first_sl= first_select();
   select_result *tmp_result;
   bool is_union;
   TABLE *empty_table= 0;
@@ -165,7 +181,7 @@ int st_select_lex_unit::prepare(THD *thd_arg, select_result *sel_result,
     if (describe)
     {
       /* fast reinit for EXPLAIN */
-      for (sl= first_select_in_union(); sl; sl= sl->next_select())
+      for (sl= first_sl; sl; sl= sl->next_select())
       {
 	sl->join->result= result;
 	select_limit_cnt= HA_POS_ERROR;
@@ -173,34 +189,35 @@ int st_select_lex_unit::prepare(THD *thd_arg, select_result *sel_result,
 	if (!sl->join->procedure &&
 	    result->prepare(sl->join->fields_list, this))
 	{
-	  DBUG_RETURN(1);
+	  DBUG_RETURN(TRUE);
 	}
 	sl->join->select_options|= SELECT_DESCRIBE;
 	sl->join->reinit();
       }
     }
-    DBUG_RETURN(0);
+    DBUG_RETURN(FALSE);
   }
   prepared= 1;
-  res= 0;
+  res= FALSE;
   
-  thd_arg->lex->current_select= sl= first_select= first_select_in_union();
-  found_rows_for_union= first_select->options & OPTION_FOUND_ROWS;
-  is_union= test(first_select->next_select());
+  thd_arg->lex->current_select= sl= first_sl;
+  found_rows_for_union= first_sl->options & OPTION_FOUND_ROWS;
+  is_union= first_sl->next_select() || fake_select_lex;
 
   /* Global option */
 
   if (is_union)
   {
-    if (!(tmp_result= union_result= new select_union(0)))
+    if (!(tmp_result= union_result= new select_union))
       goto err;
-    union_result->tmp_table_param.init();
     if (describe)
       tmp_result= sel_result;
   }
   else
     tmp_result= sel_result;
 
+  sl->context.resolve_in_select_list= TRUE;
+
   for (;sl; sl= sl->next_select())
   {
     bool can_skip_order_by;
@@ -208,22 +225,24 @@ int st_select_lex_unit::prepare(THD *thd_arg, select_result *sel_result,
     JOIN *join= new JOIN(thd_arg, sl->item_list, 
 			 sl->options | thd_arg->options | additional_options,
 			 tmp_result);
+    /*
+      setup_tables_done_option should be set only for very first SELECT,
+      because it protect from secont setup_tables call for select-like non
+      select commands (DELETE/INSERT/...) and they use only very first
+      SELECT (for union it can be only INSERT ... SELECT).
+    */
+    additional_options&= ~OPTION_SETUP_TABLES_DONE;
     if (!join)
       goto err;
 
     thd_arg->lex->current_select= sl;
-    offset_limit_cnt= sl->offset_limit;
-    select_limit_cnt= sl->select_limit+sl->offset_limit;
-    if (select_limit_cnt < sl->select_limit)
-      select_limit_cnt= HA_POS_ERROR;		// no limit
 
-    can_skip_order_by= is_union &&
-                       (!sl->braces || select_limit_cnt == HA_POS_ERROR);
+    can_skip_order_by= is_union && !(sl->braces && sl->explicit_limit);
 
     res= join->prepare(&sl->ref_pointer_array,
 		       (TABLE_LIST*) sl->table_list.first, sl->with_wild,
 		       sl->where,
-                       (can_skip_order_by ? 0 : sl->order_list.elements) +
+                      (can_skip_order_by ? 0 : sl->order_list.elements) +
                        sl->group_list.elements,
                        can_skip_order_by ?
                        (ORDER*) 0 : (ORDER *)sl->order_list.first,
@@ -235,9 +254,16 @@ int st_select_lex_unit::prepare(THD *thd_arg, select_result *sel_result,
     /* There are no * in the statement anymore (for PS) */
     sl->with_wild= 0;
     last_procedure= join->procedure;
-    if (res || thd_arg->is_fatal_error)
+
+    if ((res= (res || thd_arg->is_fatal_error)))
       goto err;
-    if (sl == first_select)
+    /*
+      Use items list of underlaid select for derived tables to preserve
+      information about fields lengths and exact types
+    */
+    if (!is_union)
+      types= first_sl->item_list;
+    else if (sl == first_sl)
     {
       /*
         We need to create an empty table object. It is used
@@ -273,7 +299,7 @@ int st_select_lex_unit::prepare(THD *thd_arg, select_result *sel_result,
       while ((type= tp++, item_tmp= it++))
       {
         if (((Item_type_holder*)type)->join_types(thd_arg, item_tmp))
-	  DBUG_RETURN(-1);
+	  DBUG_RETURN(TRUE);
       }
     }
   }
@@ -285,8 +311,9 @@ int st_select_lex_unit::prepare(THD *thd_arg, select_result *sel_result,
       all collations together for UNION.
     */
     List_iterator_fast<Item> tp(types);
-    Item_arena *arena= thd->current_arena;
     Item *type;
+    ulonglong create_options;
+
     while ((type= tp++))
     {
       if (type->result_type() == STRING_RESULT &&
@@ -296,56 +323,53 @@ int st_select_lex_unit::prepare(THD *thd_arg, select_result *sel_result,
         goto err;
       }
     }
+    
+    create_options= (first_sl->options | thd_arg->options |
+                     TMP_TABLE_ALL_COLUMNS);
+    /*
+      Force the temporary table to be a MyISAM table if we're going to use
+      fullext functions (MATCH ... AGAINST .. IN BOOLEAN MODE) when reading
+      from it (this should be removed in 5.2 when fulltext search is moved 
+      out of MyISAM).
+    */
+    if (global_parameters->ftfunc_list->elements)
+      create_options= create_options | TMP_TABLE_FORCE_MYISAM;
 
-    union_result->tmp_table_param.field_count= types.elements;
-    if (!(table= create_tmp_table(thd_arg,
-				  &union_result->tmp_table_param, types,
-				  (ORDER*) 0, (bool) union_distinct, 1, 
-				  (first_select_in_union()->options |
-				   thd_arg->options |
-				   TMP_TABLE_ALL_COLUMNS),
-				  HA_POS_ERROR, (char *) tmp_table_alias)))
+    if (union_result->create_result_table(thd, &types, test(union_distinct),
+                                          create_options, ""))
       goto err;
-    table->file->extra(HA_EXTRA_WRITE_CACHE);
-    table->file->extra(HA_EXTRA_IGNORE_DUP_KEY);
     bzero((char*) &result_table_list, sizeof(result_table_list));
     result_table_list.db= (char*) "";
-    result_table_list.real_name= result_table_list.alias= (char*) "union";
-    result_table_list.table= table;
-    union_result->set_table(table);
+    result_table_list.table_name= result_table_list.alias= (char*) "union";
+    result_table_list.table= table= union_result->table;
 
     thd_arg->lex->current_select= lex_select_save;
     if (!item_list.elements)
     {
-      /*
-        We're in statement prepare or in execution
-        of a conventional statement.
-      */
-      Item_arena *tmp_arena,backup;
-      tmp_arena= thd->change_arena_if_needed(&backup);
+      Query_arena *arena, backup_arena;
 
-      Field **field;
-      for (field= table->field; *field; field++)
+      arena= thd->activate_stmt_arena_if_needed(&backup_arena);
+      
+      res= table->fill_item_list(&item_list);
+
+      if (arena)
+        thd->restore_active_arena(arena, &backup_arena);
+
+      if (res)
+        goto err;
+
+      if (thd->stmt_arena->is_stmt_prepare())
       {
-	Item_field *item= new Item_field(*field);
-	if (!item || item_list.push_back(item))
-	{
-          if (tmp_arena)
-	    thd->restore_backup_item_arena(tmp_arena, &backup);
-	  DBUG_RETURN(-1);
-	}
-      }
-      if (tmp_arena)
-        thd->restore_backup_item_arena(tmp_arena, &backup);
-      if (arena->is_stmt_prepare())
-      {
-	/* prepare fake select to initialize it correctly */
-	(void) init_prepare_fake_select_lex(thd);
+        /* Validate the global parameters of this union */
+
+	init_prepare_fake_select_lex(thd);
+        /* Should be done only once (the only item_list per statement) */
+        DBUG_ASSERT(fake_select_lex->join == 0);
 	if (!(fake_select_lex->join= new JOIN(thd, item_list, thd->options,
 					      result)))
 	{
 	  fake_select_lex->table_list.empty();
-	  DBUG_RETURN(-1);
+	  DBUG_RETURN(TRUE);
 	}
 	fake_select_lex->item_list= item_list;
 
@@ -362,42 +386,37 @@ int st_select_lex_unit::prepare(THD *thd_arg, select_result *sel_result,
 	fake_select_lex->table_list.empty();
       }
     }
-    else if (arena->is_stmt_execute())
+    else
     {
+      DBUG_ASSERT(!thd->stmt_arena->is_conventional());
       /*
-        We're in execution of a prepared statement: reset field items
-        to point at fields from the created temporary table.
+        We're in execution of a prepared statement or stored procedure:
+        reset field items to point at fields from the created temporary table.
       */
-      List_iterator_fast<Item> it(item_list);
-      for (Field **field= table->field; *field; field++)
-      {
-        Item_field *item_field= (Item_field*) it++;
-        DBUG_ASSERT(item_field);
-        item_field->reset_field(*field);
-      }
+      table->reset_item_list(&item_list);
     }
   }
 
   thd_arg->lex->current_select= lex_select_save;
 
-  DBUG_RETURN(res || thd_arg->is_fatal_error ? 1 : 0);
+  DBUG_RETURN(res || thd_arg->is_fatal_error);
 
 err:
   thd_arg->lex->current_select= lex_select_save;
-  DBUG_RETURN(-1);
+  DBUG_RETURN(TRUE);
 }
 
 
-int st_select_lex_unit::exec()
+bool st_select_lex_unit::exec()
 {
   SELECT_LEX *lex_select_save= thd->lex->current_select;
-  SELECT_LEX *select_cursor=first_select_in_union();
+  SELECT_LEX *select_cursor=first_select();
   ulonglong add_rows=0;
   ha_rows examined_rows= 0;
   DBUG_ENTER("st_select_lex_unit::exec");
 
   if (executed && !uncacheable && !describe)
-    DBUG_RETURN(0);
+    DBUG_RETURN(FALSE);
   executed= 1;
   
   if (uncacheable || !item || !item->assigned() || describe)
@@ -425,12 +444,8 @@ int st_select_lex_unit::exec()
 	res= sl->join->reinit();
       else
       {
-	if (sl != global_parameters && !describe)
-	{
-	  offset_limit_cnt= sl->offset_limit;
-	  select_limit_cnt= sl->select_limit+sl->offset_limit;
-	}
-	else
+        set_limit(sl);
+	if (sl == global_parameters || describe)
 	{
 	  offset_limit_cnt= 0;
 	  /*
@@ -439,11 +454,7 @@ int st_select_lex_unit::exec()
 	  */
 	  if (sl->order_list.first || describe)
 	    select_limit_cnt= HA_POS_ERROR;
-	  else
-	    select_limit_cnt= sl->select_limit+sl->offset_limit;
-	}
-	if (select_limit_cnt < sl->select_limit)
-	  select_limit_cnt= HA_POS_ERROR;		// no limit
+        }
 
         /*
           When using braces, SQL_CALC_FOUND_ROWS affects the whole query:
@@ -462,11 +473,13 @@ int st_select_lex_unit::exec()
         if (sl == union_distinct)
 	{
 	  if (table->file->disable_indexes(HA_KEY_SWITCH_ALL))
-	    DBUG_RETURN(1);
+	    DBUG_RETURN(TRUE);
 	  table->no_keyread=1;
 	}
 	res= sl->join->error;
-	offset_limit_cnt= sl->offset_limit;
+	offset_limit_cnt= (ha_rows)(sl->offset_limit ?
+                                    sl->offset_limit->val_uint() :
+                                    0);
 	if (!res)
 	{
 	  examined_rows+= thd->examined_row_count;
@@ -483,7 +496,12 @@ int st_select_lex_unit::exec()
 	DBUG_RETURN(res);
       }
       /* Needed for the following test and for records_at_start in next loop */
-      table->file->info(HA_STATUS_VARIABLE);
+      int error= table->file->info(HA_STATUS_VARIABLE);
+      if(error)
+      {
+        table->file->print_error(error, MYF(0));
+        DBUG_RETURN(1);
+      }
       if (found_rows_for_union && !sl->braces && 
           select_limit_cnt != HA_POS_ERROR)
       {
@@ -501,14 +519,15 @@ int st_select_lex_unit::exec()
   optimized= 1;
 
   /* Send result to 'result' */
-  res= -1;
+  res= TRUE;
   {
     List<Item_func_match> empty_list;
     empty_list.empty();
 
     if (!thd->is_fatal_error)				// Check if EOM
     {
-      ulong options_tmp= init_prepare_fake_select_lex(thd);
+      set_limit(global_parameters);
+      init_prepare_fake_select_lex(thd);
       JOIN *join= fake_select_lex->join;
       if (!join)
       {
@@ -516,11 +535,11 @@ int st_select_lex_unit::exec()
 	  allocate JOIN for fake select only once (prevent
 	  mysql_select automatic allocation)
 	*/
-	if (!(fake_select_lex->join= new JOIN(thd, item_list, thd->options,
-					      result)))
+	if (!(fake_select_lex->join= new JOIN(thd, item_list,
+					      fake_select_lex->options, result)))
 	{
 	  fake_select_lex->table_list.empty();
-	  DBUG_RETURN(-1);
+	  DBUG_RETURN(TRUE);
 	}
 
 	/*
@@ -532,12 +551,14 @@ int st_select_lex_unit::exec()
       else
       {
 	JOIN_TAB *tab,*end;
-	for (tab=join->join_tab,end=tab+join->tables ; tab != end ; tab++)
+	for (tab=join->join_tab, end=tab+join->tables ;
+	     tab && tab != end ;
+	     tab++)
 	{
 	  delete tab->select;
 	  delete tab->quick;
 	}
-	join->init(thd, item_list, thd->options, result);
+	join->init(thd, item_list, fake_select_lex->options, result);
       }
       res= mysql_select(thd, &fake_select_lex->ref_pointer_array,
 			&result_table_list,
@@ -545,7 +566,7 @@ int st_select_lex_unit::exec()
 			global_parameters->order_list.elements,
 			(ORDER*)global_parameters->order_list.first,
 			(ORDER*) NULL, NULL, (ORDER*) NULL,
-			options_tmp | SELECT_NO_UNLOCK,
+			fake_select_lex->options | SELECT_NO_UNLOCK,
 			result, this, fake_select_lex);
 
       fake_select_lex->table_list.empty();
@@ -565,14 +586,14 @@ int st_select_lex_unit::exec()
 }
 
 
-int st_select_lex_unit::cleanup()
+bool st_select_lex_unit::cleanup()
 {
   int error= 0;
   DBUG_ENTER("st_select_lex_unit::cleanup");
 
   if (cleaned)
   {
-    DBUG_RETURN(0);
+    DBUG_RETURN(FALSE);
   }
   cleaned= 1;
 
@@ -584,33 +605,21 @@ int st_select_lex_unit::cleanup()
       free_tmp_table(thd, table);
     table= 0; // Safety
   }
-  JOIN *join;
-  SELECT_LEX *sl= first_select_in_union();
-  for (; sl; sl= sl->next_select())
+
+  for (SELECT_LEX *sl= first_select(); sl; sl= sl->next_select())
+    error|= sl->cleanup();
+
+  if (fake_select_lex)
   {
-    if ((join= sl->join))
+    JOIN *join;
+    if ((join= fake_select_lex->join))
     {
-      error|= sl->join->cleanup();
-      delete join;
-    }
-    else
-    {
-      // it can be DO/SET with subqueries
-      for (SELECT_LEX_UNIT *lex_unit= sl->first_inner_unit();
-	   lex_unit != 0;
-	   lex_unit= lex_unit->next_unit())
-      {
-	error|= lex_unit->cleanup();
-      }
+      join->tables_list= 0;
+      join->tables= 0;
     }
+    error|= fake_select_lex->cleanup();
   }
-  if (fake_select_lex && (join= fake_select_lex->join))
-  {
-    join->tables_list= 0;
-    join->tables= 0;
-    error|= join->cleanup();
-    delete join;
-  }
+
   DBUG_RETURN(error);
 }
 
@@ -646,21 +655,98 @@ void st_select_lex_unit::reinit_exec_mechanism()
     old_result	old select_result object
 
   RETURN
-    0 - OK
-    -1 - error
+    FALSE - OK
+    TRUE  - error
 */
 
-int st_select_lex_unit::change_result(select_subselect *result,
-				      select_subselect *old_result)
+bool st_select_lex_unit::change_result(select_subselect *result,
+                                       select_subselect *old_result)
 {
-  int res= 0;
-  for (SELECT_LEX *sl= first_select_in_union(); sl; sl= sl->next_select())
+  bool res= FALSE;
+  for (SELECT_LEX *sl= first_select(); sl; sl= sl->next_select())
   {
     if (sl->join && sl->join->result == old_result)
-      if ((res= sl->join->change_result(result)))
-	return (res);
+      if (sl->join->change_result(result))
+	return TRUE;
   }
   if (fake_select_lex && fake_select_lex->join)
     res= fake_select_lex->join->change_result(result);
   return (res);
 }
+
+/*
+  Get column type information for this unit.
+
+  SYNOPSIS
+    st_select_lex_unit::get_unit_column_types()
+
+  DESCRIPTION
+    For a single-select the column types are taken
+    from the list of selected items. For a union this function
+    assumes that st_select_lex_unit::prepare has been called
+    and returns the type holders that were created for unioned
+    column types of all selects.
+
+  NOTES
+    The implementation of this function should be in sync with
+    st_select_lex_unit::prepare()
+*/
+
+List<Item> *st_select_lex_unit::get_unit_column_types()
+{
+  SELECT_LEX *sl= first_select();
+  bool is_union= test(sl->next_select());
+  bool is_procedure= test(sl->join->procedure);
+
+  if (is_procedure)
+  {
+    /* Types for "SELECT * FROM t1 procedure analyse()"
+       are generated during execute */
+    return &sl->join->procedure_fields_list;
+  }
+
+
+  if (is_union)
+  {
+    DBUG_ASSERT(prepared);
+    /* Types are generated during prepare */
+    return &types;
+  }
+
+  return &sl->item_list;
+}
+
+bool st_select_lex::cleanup()
+{
+  bool error= FALSE;
+  DBUG_ENTER("st_select_lex::cleanup()");
+
+  if (join)
+  {
+    DBUG_ASSERT((st_select_lex*)join->select_lex == this);
+    error= join->destroy();
+    delete join;
+    join= 0;
+  }
+  for (SELECT_LEX_UNIT *lex_unit= first_inner_unit(); lex_unit ;
+       lex_unit= lex_unit->next_unit())
+  {
+    error= (bool) ((uint) error | (uint) lex_unit->cleanup());
+  }
+  DBUG_RETURN(error);
+}
+
+
+void st_select_lex::cleanup_all_joins(bool full)
+{
+  SELECT_LEX_UNIT *unit;
+  SELECT_LEX *sl;
+
+  if (join)
+    join->cleanup(full);
+
+  for (unit= first_inner_unit(); unit; unit= unit->next_unit())
+    for (sl= unit->first_select(); sl; sl= sl->next_select())
+      sl->cleanup_all_joins(full);
+}
+
