@@ -813,7 +813,7 @@ end:
     printf("Write records: %d\nUpdate records: %d\nSame-key-read: %d\nDelete records: %d\n", write_count,update,dupp_keys,opt_delete);
     if (rec_pointer_size)
       printf("Record pointer size:  %d\n",rec_pointer_size);
-    printf("myisam_block_size:    %u\n", myisam_block_size);
+    printf("myisam_block_size:    %lu\n", myisam_block_size);
     if (key_cacheing)
     {
       puts("Key cache used");
@@ -831,17 +831,19 @@ end:
       puts("Locking used");
     if (use_blob)
       puts("blobs used");
-#if 0
     printf("key cache status: \n\
 blocks used:%10lu\n\
+not flushed:%10lu\n\
 w_requests: %10lu\n\
 writes:     %10lu\n\
 r_requests: %10lu\n\
 reads:      %10lu\n",
-	   my_blocks_used,
-           my_cache_w_requests, my_cache_write,
-	   my_cache_r_requests, my_cache_read);
-#endif
+           dflt_key_cache->blocks_used,
+           dflt_key_cache->global_blocks_changed,
+           (ulong) dflt_key_cache->global_cache_w_requests,
+           (ulong) dflt_key_cache->global_cache_write,
+           (ulong) dflt_key_cache->global_cache_r_requests,
+           (ulong) dflt_key_cache->global_cache_read);
   }
   end_key_cache(dflt_key_cache,1);
   if (blob_buffer)
