@@ -52,6 +52,9 @@ int mi_write(MI_INFO *info, byte *record)
   DBUG_ENTER("mi_write");
   DBUG_PRINT("enter",("isam: %d  data: %d",info->s->kfile,info->dfile));
 
+  DBUG_EXECUTE_IF("myisam_pretend_crashed_table_on_usage",
+                  mi_print_error(info->s, HA_ERR_CRASHED);
+                  DBUG_RETURN(my_errno= HA_ERR_CRASHED););
   if (share->options & HA_OPTION_READ_ONLY_DATA)
   {
     DBUG_RETURN(my_errno=EACCES);
@@ -64,7 +67,8 @@ int mi_write(MI_INFO *info, byte *record)
 			       MYF(MY_SEEK_NOT_DONE) | info->lock_wait))
     goto err;
 #endif
-  filepos= ((share->state.dellink != HA_OFFSET_ERROR) ?
+  filepos= ((share->state.dellink != HA_OFFSET_ERROR &&
+             !info->append_insert_at_end) ?
 	    share->state.dellink :
 	    info->state->data_file_length);
 
@@ -97,7 +101,7 @@ int mi_write(MI_INFO *info, byte *record)
   buff=info->lastkey2;
   for (i=0 ; i < share->base.keys ; i++)
   {
-    if (((ulonglong) 1 << i) & share->state.key_map)
+    if (mi_is_key_active(share->state.key_map, i))
     {
       bool local_lock_tree= (lock_tree &&
 			     !(info->bulk_insert &&
@@ -128,6 +132,10 @@ int mi_write(MI_INFO *info, byte *record)
           goto err;
         }
       }
+
+      /* The above changed info->lastkey2. Inform mi_rnext_same(). */
+      info->update&= ~HA_STATE_RNEXT_SAME;
+
       if (local_lock_tree)
         rw_unlock(&share->key_root_lock[i]);
     }
@@ -138,10 +146,11 @@ int mi_write(MI_INFO *info, byte *record)
   {
     if ((*share->write_record)(info,record))
       goto err;
-    share->state.checksum+=info->checksum;
+    info->state->checksum+=info->checksum;
   }
   if (share->base.auto_key)
-    update_auto_increment(info,record);
+    set_if_bigger(info->s->state.auto_increment,
+                  retrieve_auto_increment(info, record));
   info->update= (HA_STATE_CHANGED | HA_STATE_AKTIV | HA_STATE_WRITTEN |
 		 HA_STATE_ROW_CHANGED);
   info->state->records++;
@@ -171,7 +180,7 @@ err:
     info->errkey= (int) i;
     while ( i-- > 0)
     {
-      if (((ulonglong) 1 << i) & share->state.key_map)
+      if (mi_is_key_active(share->state.key_map, i))
       {
 	bool local_lock_tree= (lock_tree &&
 			       !(info->bulk_insert &&
@@ -203,7 +212,10 @@ err:
     }
   }
   else
+  {
+    mi_print_error(info->s, HA_ERR_CRASHED);
     mi_mark_crashed(info);
+  }
   info->update= (HA_STATE_CHANGED | HA_STATE_WRITTEN | HA_STATE_ROW_CHANGED);
   my_errno=save_errno;
 err2:
@@ -249,7 +261,7 @@ int _mi_ck_write_btree(register MI_INFO *info, uint keynr, uchar *key,
     comp_flag=SEARCH_BIGGER;			/* Put after same key */
   else if (keyinfo->flag & (HA_NOSAME|HA_FULLTEXT))
   {
-    comp_flag=SEARCH_FIND | SEARCH_UPDATE;	/* No dupplicates */
+    comp_flag=SEARCH_FIND | SEARCH_UPDATE;	/* No duplicates */
     if (keyinfo->flag & HA_NULL_ARE_EQUAL)
       comp_flag|= SEARCH_NULL_ARE_EQUAL;
   }
@@ -327,7 +339,7 @@ static int w_search(register MI_INFO *info, register MI_KEYDEF *keyinfo,
   my_bool was_last_key;
   my_off_t next_page, dupp_key_pos;
   DBUG_ENTER("w_search");
-  DBUG_PRINT("enter",("page: %ld",page));
+  DBUG_PRINT("enter",("page: %ld", (long) page));
 
   search_key_length= (comp_flag & SEARCH_FIND) ? key_length : USE_WHOLE_KEY;
   if (!(temp_buff= (uchar*) my_alloca((uint) keyinfo->block_length+
@@ -348,6 +360,7 @@ static int w_search(register MI_INFO *info, register MI_KEYDEF *keyinfo,
       dupp_key_pos=_mi_dpos(info,0,keybuff+tmp_key_length);
     else
       dupp_key_pos= HA_OFFSET_ERROR;
+
     if (keyinfo->flag & HA_FULLTEXT)
     {
       uint off;
@@ -414,8 +427,30 @@ err:
 } /* w_search */
 
 
-	/* Insert new key at right of key_pos */
-	/* Returns 2 if key contains key to upper level */
+/*
+  Insert new key.
+
+  SYNOPSIS
+    _mi_insert()
+    info                        Open table information.
+    keyinfo                     Key definition information.
+    key                         New key.
+    anc_buff                    Key page (beginning).
+    key_pos                     Position in key page where to insert.
+    key_buff                    Copy of previous key.
+    father_buff                 parent key page for balancing.
+    father_key_pos              position in parent key page for balancing.
+    father_page                 position of parent key page in file.
+    insert_last                 If to append at end of page.
+
+  DESCRIPTION
+    Insert new key at right of key_pos.
+
+  RETURN
+    2           if key contains key to upper level.
+    0           OK.
+    < 0         Error.
+*/
 
 int _mi_insert(register MI_INFO *info, register MI_KEYDEF *keyinfo,
 	       uchar *key, uchar *anc_buff, uchar *key_pos, uchar *key_buff,
@@ -427,7 +462,7 @@ int _mi_insert(register MI_INFO *info, register MI_KEYDEF *keyinfo,
   uchar *endpos, *prev_key;
   MI_KEY_PARAM s_temp;
   DBUG_ENTER("_mi_insert");
-  DBUG_PRINT("enter",("key_pos: %lx",key_pos));
+  DBUG_PRINT("enter",("key_pos: 0x%lx", (long) key_pos));
   DBUG_EXECUTE("key",_mi_print_key(DBUG_FILE,keyinfo->seg,key,USE_WHOLE_KEY););
 
   nod_flag=mi_test_if_nod(anc_buff);
@@ -448,14 +483,15 @@ int _mi_insert(register MI_INFO *info, register MI_KEYDEF *keyinfo,
   {
     DBUG_PRINT("test",("t_length: %d  ref_len: %d",
 		       t_length,s_temp.ref_length));
-    DBUG_PRINT("test",("n_ref_len: %d  n_length: %d  key_pos: %lx",
-		       s_temp.n_ref_length,s_temp.n_length,s_temp.key));
+    DBUG_PRINT("test",("n_ref_len: %d  n_length: %d  key_pos: 0x%lx",
+		       s_temp.n_ref_length,s_temp.n_length, (long) s_temp.key));
   }
 #endif
   if (t_length > 0)
   {
     if (t_length >= keyinfo->maxlength*2+MAX_POINTER_LENGTH)
     {
+      mi_print_error(info->s, HA_ERR_CRASHED);
       my_errno=HA_ERR_CRASHED;
       DBUG_RETURN(-1);
     }
@@ -465,6 +501,7 @@ int _mi_insert(register MI_INFO *info, register MI_KEYDEF *keyinfo,
   {
     if (-t_length >= keyinfo->maxlength*2+MAX_POINTER_LENGTH)
     {
+      mi_print_error(info->s, HA_ERR_CRASHED);
       my_errno=HA_ERR_CRASHED;
       DBUG_RETURN(-1);
     }
@@ -560,6 +597,7 @@ int _mi_split_page(register MI_INFO *info, register MI_KEYDEF *keyinfo,
 			      &after_key);
   if (!key_pos)
     DBUG_RETURN(-1);
+
   length=(uint) (key_pos-buff);
   a_length=mi_getint(buff);
   mi_putint(buff,length,nod_flag);
@@ -580,6 +618,7 @@ int _mi_split_page(register MI_INFO *info, register MI_KEYDEF *keyinfo,
 	/* Store new page */
   if (!(*keyinfo->get_key)(keyinfo,nod_flag,&key_pos,key_buff))
     DBUG_RETURN(-1);
+
   t_length=(*keyinfo->pack_key)(keyinfo,nod_flag,(uchar *) 0,
 				(uchar*) 0, (uchar*) 0,
 				key_buff, &s_temp);
@@ -638,7 +677,8 @@ uchar *_mi_find_half_pos(uint nod_flag, MI_KEYDEF *keyinfo, uchar *page,
   } while (page < end);
   *return_key_length=length;
   *after_key=page;
-  DBUG_PRINT("exit",("returns: %lx  page: %lx  half: %lx",lastpos,page,end));
+  DBUG_PRINT("exit",("returns: 0x%lx  page: 0x%lx  half: 0x%lx",
+                     (long) lastpos, (long) page, (long) end));
   DBUG_RETURN(lastpos);
 } /* _mi_find_half_pos */
 
@@ -686,13 +726,15 @@ static uchar *_mi_find_last_pos(MI_KEYDEF *keyinfo, uchar *page,
     memcpy(key, key_buff, length);		/* previous key */
     if (!(length=(*keyinfo->get_key)(keyinfo,0,&page,key_buff)))
     {
+      mi_print_error(keyinfo->share, HA_ERR_CRASHED);
       my_errno=HA_ERR_CRASHED;
       DBUG_RETURN(0);
     }
   }
   *return_key_length=last_length;
   *after_key=lastpos;
-  DBUG_PRINT("exit",("returns: %lx  page: %lx  end: %lx",prevpos,page,end));
+  DBUG_PRINT("exit",("returns: 0x%lx  page: 0x%lx  end: 0x%lx",
+                     (long) prevpos,(long) page,(long) end));
   DBUG_RETURN(prevpos);
 } /* _mi_find_last_pos */
 
@@ -728,7 +770,7 @@ static int _mi_balance_page(register MI_INFO *info, MI_KEYDEF *keyinfo,
     next_page= _mi_kpos(info->s->base.key_reflength,
 			father_key_pos+father_keylength);
     buff=info->buff;
-    DBUG_PRINT("test",("use right page: %lu",next_page));
+    DBUG_PRINT("test",("use right page: %lu", (ulong) next_page));
   }
   else
   {
@@ -737,7 +779,7 @@ static int _mi_balance_page(register MI_INFO *info, MI_KEYDEF *keyinfo,
     next_page= _mi_kpos(info->s->base.key_reflength,father_key_pos);
 					/* Fix that curr_buff is to left */
     buff=curr_buff; curr_buff=info->buff;
-    DBUG_PRINT("test",("use left page: %lu",next_page));
+    DBUG_PRINT("test",("use left page: %lu", (ulong) next_page));
   }					/* father_key_pos ptr to parting key */
 
   if (!_mi_fetch_keypage(info,keyinfo,next_page,DFLT_INIT_HITS,info->buff,0))
@@ -863,10 +905,10 @@ int _mi_ck_write_tree(register MI_INFO *info, uint keynr, uchar *key,
 
 static int keys_compare(bulk_insert_param *param, uchar *key1, uchar *key2)
 {
-  uint not_used;
+  uint not_used[2];
   return ha_key_cmp(param->info->s->keyinfo[param->keynr].seg,
-		     key1, key2, USE_WHOLE_KEY, SEARCH_SAME,
-		     &not_used);
+                    key1, key2, USE_WHOLE_KEY, SEARCH_SAME,
+                    not_used);
 }
 
 
@@ -909,20 +951,21 @@ int mi_init_bulk_insert(MI_INFO *info, ulong cache_size, ha_rows rows)
   MI_KEYDEF *key=share->keyinfo;
   bulk_insert_param *params;
   uint i, num_keys, total_keylength;
-  ulonglong key_map=0;
+  ulonglong key_map;
   DBUG_ENTER("_mi_init_bulk_insert");
   DBUG_PRINT("enter",("cache_size: %lu", cache_size));
 
   DBUG_ASSERT(!info->bulk_insert &&
 	      (!rows || rows >= MI_MIN_ROWS_TO_USE_BULK_INSERT));
 
+  mi_clear_all_keys_active(key_map);
   for (i=total_keylength=num_keys=0 ; i < share->base.keys ; i++)
   {
-    if (!(key[i].flag & HA_NOSAME) && share->base.auto_key != i+1
-        && test(share->state.key_map & ((ulonglong) 1 << i)))
+    if (! (key[i].flag & HA_NOSAME) && (share->base.auto_key != i + 1) &&
+        mi_is_key_active(share->state.key_map, i))
     {
       num_keys++;
-      key_map |=((ulonglong) 1 << i);
+      mi_set_key_active(key_map, i);
       total_keylength+=key[i].maxlength+TREE_ELEMENT_EXTRA_SIZE;
     }
   }
@@ -946,7 +989,7 @@ int mi_init_bulk_insert(MI_INFO *info, ulong cache_size, ha_rows rows)
   params=(bulk_insert_param *)(info->bulk_insert+share->base.keys);
   for (i=0 ; i < share->base.keys ; i++)
   {
-    if (test(key_map & ((ulonglong) 1 << i)))
+    if (mi_is_key_active(key_map, i))
     {
       params->info=info;
       params->keynr=i;
