@@ -28,7 +28,7 @@
 LIST *list_add(LIST *root, LIST *element)
 {
   DBUG_ENTER("list_add");
-  DBUG_PRINT("enter",("root: 0x%lx  element: %lx", root, element));
+  DBUG_PRINT("enter",("root: 0x%lx  element: 0x%lx", (long) root, (long) element));
   if (root)
   {
     if (root->prev)			/* If add in mid of list */
@@ -109,7 +109,7 @@ int list_walk(LIST *list, list_walk_action action, gptr argument)
   {
     if ((error = (*action)(list->data,argument)))
       return error;
-    list=rest(list);
+    list=list_rest(list);
   }
   return 0;
 }
