@@ -501,7 +501,7 @@ MY_DIR	*my_dir(const char *path, myf MyFlags)
   if (!(MyFlags & MY_DONT_SORT))
     qsort((void *) result->dir_entry, result->number_off_files,
           sizeof(FILEINFO), (qsort_cmp) comp_names);
-  DBUG_PRINT(exit, ("found %d files", result->number_off_files));
+  DBUG_PRINT("exit", ("found %d files", result->number_off_files));
   DBUG_RETURN(result);
 error:
   my_errno=errno;
@@ -638,7 +638,7 @@ MY_STAT *my_stat(const char *path, MY_STAT *stat_area, myf my_flags)
   int m_used;
   DBUG_ENTER("my_stat");
   DBUG_PRINT("my", ("path: '%s', stat_area: 0x%lx, MyFlags: %d", path,
-	     (byte *) stat_area, my_flags));
+                    (long) stat_area, my_flags));
 
   if ((m_used= (stat_area == NULL)))
     if (!(stat_area = (MY_STAT *) my_malloc(sizeof(MY_STAT), my_flags)))
