@@ -30,7 +30,10 @@ pthread_key(struct st_my_thread_var, THR_KEY_mysys);
 #endif /* USE_TLS */
 pthread_mutex_t THR_LOCK_malloc,THR_LOCK_open,
 	        THR_LOCK_lock,THR_LOCK_isam,THR_LOCK_myisam,THR_LOCK_heap,
-	        THR_LOCK_net, THR_LOCK_charset; 
+                THR_LOCK_net, THR_LOCK_charset, THR_LOCK_threads;
+pthread_cond_t  THR_COND_threads;
+uint            THR_thread_count= 0;
+uint 		my_thread_end_wait_time= 5;
 #if !defined(HAVE_LOCALTIME_R) || !defined(HAVE_GMTIME_R)
 pthread_mutex_t LOCK_localtime_r;
 #endif
@@ -41,7 +44,7 @@ pthread_mutex_t LOCK_gethostbyname_r;
 pthread_mutexattr_t my_fast_mutexattr;
 #endif
 #ifdef PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP
-pthread_mutexattr_t my_errchk_mutexattr;
+pthread_mutexattr_t my_errorcheck_mutexattr;
 #endif
 
 /*
@@ -62,19 +65,28 @@ my_bool my_thread_global_init(void)
     fprintf(stderr,"Can't initialize threads: error %d\n",errno);
     return 1;
   }
+
 #ifdef PTHREAD_ADAPTIVE_MUTEX_INITIALIZER_NP
-  pthread_mutexattr_init(&my_fast_mutexattr);
   /*
-    Note that the following statement may give a compiler warning under
-    some configurations, but there isn't anything we can do about this as
-    this is a bug in the header files for the thread implementation
+    Set mutex type to "fast" a.k.a "adaptive"
+
+    In this case the thread may steal the mutex from some other thread
+    that is waiting for the same mutex.  This will save us some
+    context switches but may cause a thread to 'starve forever' while
+    waiting for the mutex (not likely if the code within the mutex is
+    short).
   */
-  pthread_mutexattr_setkind_np(&my_fast_mutexattr,PTHREAD_MUTEX_ADAPTIVE_NP);
+  pthread_mutexattr_init(&my_fast_mutexattr);
+  pthread_mutexattr_settype(&my_fast_mutexattr,
+                            PTHREAD_MUTEX_ADAPTIVE_NP);
 #endif
 #ifdef PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP
-  pthread_mutexattr_init(&my_errchk_mutexattr);
-  pthread_mutexattr_setkind_np(&my_errchk_mutexattr,
-			       PTHREAD_MUTEX_ERRORCHECK_NP);
+  /*
+    Set mutex type to "errorcheck"
+  */
+  pthread_mutexattr_init(&my_errorcheck_mutexattr);
+  pthread_mutexattr_settype(&my_errorcheck_mutexattr,
+                            PTHREAD_MUTEX_ERRORCHECK);
 #endif
 
   pthread_mutex_init(&THR_LOCK_malloc,MY_MUTEX_INIT_FAST);
@@ -85,6 +97,8 @@ my_bool my_thread_global_init(void)
   pthread_mutex_init(&THR_LOCK_heap,MY_MUTEX_INIT_FAST);
   pthread_mutex_init(&THR_LOCK_net,MY_MUTEX_INIT_FAST);
   pthread_mutex_init(&THR_LOCK_charset,MY_MUTEX_INIT_FAST);
+  pthread_mutex_init(&THR_LOCK_threads,MY_MUTEX_INIT_FAST);
+  pthread_cond_init(&THR_COND_threads, NULL);
 #if defined( __WIN__) || defined(OS2)
   win_pthread_init();
 #endif
@@ -105,12 +119,33 @@ my_bool my_thread_global_init(void)
 
 void my_thread_global_end(void)
 {
+  struct timespec abstime;
+  my_bool all_threads_killed= 1;
+
+  set_timespec(abstime, my_thread_end_wait_time);
+  pthread_mutex_lock(&THR_LOCK_threads);
+  while (THR_thread_count > 0)
+  {
+    int error= pthread_cond_timedwait(&THR_COND_threads, &THR_LOCK_threads,
+                                      &abstime);
+    if (error == ETIMEDOUT || error == ETIME)
+    {
+      if (THR_thread_count)
+        fprintf(stderr,
+                "Error in my_thread_global_end(): %d threads didn't exit\n",
+                THR_thread_count);
+      all_threads_killed= 0;
+      break;
+    }
+  }
+  pthread_mutex_unlock(&THR_LOCK_threads);
+
   pthread_key_delete(THR_KEY_mysys);
 #ifdef PTHREAD_ADAPTIVE_MUTEX_INITIALIZER_NP
   pthread_mutexattr_destroy(&my_fast_mutexattr);
 #endif
 #ifdef PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP
-  pthread_mutexattr_destroy(&my_errchk_mutexattr);
+  pthread_mutexattr_destroy(&my_errorcheck_mutexattr);
 #endif
   pthread_mutex_destroy(&THR_LOCK_malloc);
   pthread_mutex_destroy(&THR_LOCK_open);
@@ -120,6 +155,11 @@ void my_thread_global_end(void)
   pthread_mutex_destroy(&THR_LOCK_heap);
   pthread_mutex_destroy(&THR_LOCK_net);
   pthread_mutex_destroy(&THR_LOCK_charset);
+  if (all_threads_killed)
+  {
+    pthread_mutex_destroy(&THR_LOCK_threads);
+    pthread_cond_destroy (&THR_COND_threads);
+  }
 #if !defined(HAVE_LOCALTIME_R) || !defined(HAVE_GMTIME_R)
   pthread_mutex_destroy(&LOCK_localtime_r);
 #endif
@@ -131,10 +171,23 @@ void my_thread_global_end(void)
 static long thread_id=0;
 
 /*
-  We can't use mutex_locks here if we are using windows as
-  we may have compiled the program with SAFE_MUTEX, in which
-  case the checking of mutex_locks will not work until
-  the pthread_self thread specific variable is initialized.
+  Allocate thread specific memory for the thread, used by mysys and dbug
+
+  SYNOPSIS
+    my_thread_init()
+
+  NOTES
+    We can't use mutex_locks here if we are using windows as
+    we may have compiled the program with SAFE_MUTEX, in which
+    case the checking of mutex_locks will not work until
+    the pthread_self thread specific variable is initialized.
+
+   This function may called multiple times for a thread, for example
+   if one uses my_init() followed by mysql_server_init().
+
+  RETURN
+    0  ok
+    1  Fatal error; mysys/dbug functions can't be used
 */
 
 my_bool my_thread_init(void)
@@ -145,9 +198,6 @@ my_bool my_thread_init(void)
 #ifdef EXTRA_DEBUG_THREADS
   fprintf(stderr,"my_thread_init(): thread_id=%ld\n",pthread_self());
 #endif  
-#if !defined(__WIN__) || defined(USE_TLS) || ! defined(SAFE_MUTEX)
-  pthread_mutex_lock(&THR_LOCK_lock);
-#endif
 
 #if !defined(__WIN__) || defined(USE_TLS)
   if (my_pthread_getspecific(struct st_my_thread_var *,THR_KEY_mysys))
@@ -165,7 +215,7 @@ my_bool my_thread_init(void)
   }
   pthread_setspecific(THR_KEY_mysys,tmp);
 
-#else
+#else /* defined(__WIN__) && !(defined(USE_TLS) */
   /*
     Skip initialization if the thread specific variable is already initialized
   */
@@ -173,7 +223,6 @@ my_bool my_thread_init(void)
     goto end;
   tmp= &THR_KEY_mysys;
 #endif
-  tmp->id= ++thread_id;
 #if defined(__WIN__) && defined(EMBEDDED_LIBRARY)
   tmp->thread_self= (pthread_t)getpid();
 #endif
@@ -181,22 +230,35 @@ my_bool my_thread_init(void)
   pthread_cond_init(&tmp->suspend, NULL);
   tmp->init= 1;
 
+  pthread_mutex_lock(&THR_LOCK_threads);
+  tmp->id= ++thread_id;
+  ++THR_thread_count;
+  pthread_mutex_unlock(&THR_LOCK_threads);
 end:
-#if !defined(__WIN__) || defined(USE_TLS) || ! defined(SAFE_MUTEX)
-  pthread_mutex_unlock(&THR_LOCK_lock);
-#endif
   return error;
 }
 
 
+/*
+  Deallocate memory used by the thread for book-keeping
+
+  SYNOPSIS
+    my_thread_end()
+
+  NOTE
+    This may be called multiple times for a thread.
+    This happens for example when one calls 'mysql_server_init()'
+    mysql_server_end() and then ends with a mysql_end().
+*/
+
 void my_thread_end(void)
 {
   struct st_my_thread_var *tmp;
   tmp= my_pthread_getspecific(struct st_my_thread_var*,THR_KEY_mysys);
 
 #ifdef EXTRA_DEBUG_THREADS
-  fprintf(stderr,"my_thread_end(): tmp=%p,thread_id=%ld\n",
-	  tmp,pthread_self());
+  fprintf(stderr,"my_thread_end(): tmp: 0x%lx  thread_id=%ld\n",
+	  (long) tmp, pthread_self());
 #endif  
   if (tmp && tmp->init)
   {
@@ -218,6 +280,18 @@ void my_thread_end(void)
 #else
     tmp->init= 0;
 #endif
+
+    /*
+      Decrement counter for number of running threads. We are using this
+      in my_thread_global_end() to wait until all threads have called
+      my_thread_end and thus freed all memory they have allocated in
+      my_thread_init() and DBUG_xxxx
+    */
+    pthread_mutex_lock(&THR_LOCK_threads);
+    DBUG_ASSERT(THR_thread_count != 0);
+    if (--THR_thread_count == 0)
+      pthread_cond_signal(&THR_COND_threads);
+   pthread_mutex_unlock(&THR_LOCK_threads);
   }
   /* The following free has to be done, even if my_thread_var() is 0 */
 #if (!defined(__WIN__) && !defined(OS2)) || defined(USE_TLS)
