@@ -192,13 +192,13 @@ public:
   enum ErrorCode {
     NoError = 0,
     Busy = 701,
+    BusyWithNR = 711,
     NotMaster = 702,
     TriggerNotFound = 4238,
     TriggerExists = 4239,
     IndexNameTooLong = 4241,
     TooManyIndexes = 4242,
     IndexExists = 4244,
-    AttributeNotStored = 4245,
     AttributeNullable = 4246,
     BadRequestType = 4247,
     InvalidName = 4248,
@@ -207,7 +207,7 @@ public:
     NotUnique = 4251,
     AllocationError = 4252,
     CreateIndexTableFailed = 4253,
-    InvalidAttributeOrder = 4255
+    DuplicateAttributes = 4258
   };
 
   CreateIndxConf m_conf;
