@@ -46,6 +46,7 @@ public:
 protected:
 
   void execSTTOR(Signal* signal);
+  void execREAD_CONFIG_REQ(Signal* signal);
   void execDUMP_STATE_ORD(Signal* signal);
   void execREAD_NODESCONF(Signal* signal);
   void execNODE_FAILREP(Signal* signal);
@@ -67,6 +68,7 @@ protected:
   void execBACKUP_DATA(Signal* signal);
   void execSTART_BACKUP_REQ(Signal* signal);
   void execBACKUP_FRAGMENT_REQ(Signal* signal);
+  void execBACKUP_FRAGMENT_COMPLETE_REP(Signal* signal);
   void execSTOP_BACKUP_REQ(Signal* signal);
   void execBACKUP_STATUS_REQ(Signal* signal);
   void execABORT_BACKUP_ORD(Signal* signal);
@@ -182,10 +184,12 @@ public:
   typedef Ptr<Attribute> AttributePtr;
   
   struct Fragment {
+    Uint64 noOfRecords;
     Uint32 tableId;
-    Uint32 node;
-    Uint16 scanned;  // 0 = not scanned x = scanned by node x
-    Uint16 scanning; // 0 = not scanning x = scanning on node x
+    Uint8  node;
+    Uint8  scanned;  // 0 = not scanned x = scanned by node x
+    Uint8  scanning; // 0 = not scanning x = scanning on node x
+    Uint8  unused1;
     Uint32 nextPool;
   };
   typedef Ptr<Fragment> FragmentPtr;
@@ -193,9 +197,10 @@ public:
   struct Table {
     Table(ArrayPool<Attribute> &, ArrayPool<Fragment> &);
     
+    Uint64 noOfRecords;
+
     Uint32 tableId;
     Uint32 schemaVersion;
-    Uint32 frag_mask;
     Uint32 tableType;
     Uint32 noOfNull;
     Uint32 noOfAttributes;
@@ -269,8 +274,8 @@ public:
     Uint32 tablePtr;    // Ptr.i to current table
 
     FsBuffer dataBuffer;
-    Uint32 noOfRecords;
-    Uint32 noOfBytes;
+    Uint64 noOfRecords;
+    Uint64 noOfBytes;
     Uint32 maxRecordSize;
     
   private:
@@ -413,6 +418,7 @@ public:
     
     Uint32 clientRef;
     Uint32 clientData;
+    Uint32 flags;
     Uint32 backupId;
     Uint32 backupKey[2];
     Uint32 masterRef;
@@ -512,8 +518,10 @@ public:
   Config c_defaults;
   Uint32 m_diskless;
 
-  STATIC_CONST(NO_OF_PAGES_META_FILE = 2);
-
+  STATIC_CONST(NO_OF_PAGES_META_FILE = 
+	       (MAX_WORDS_META_FILE + BACKUP_WORDS_PER_PAGE - 1) / 
+	       BACKUP_WORDS_PER_PAGE);
+  
   /**
    * Pools
    */
@@ -526,8 +534,6 @@ public:
   ArrayPool<Node> c_nodePool;
   ArrayPool<TriggerRecord> c_triggerPool;
 
-  Uint32 calculate_frag_mask(Uint32);
-
   void checkFile(Signal*, BackupFilePtr);
   void checkScan(Signal*, BackupFilePtr);
   void fragmentCompleted(Signal*, BackupFilePtr);
@@ -595,7 +601,7 @@ public:
   
   bool insertFileHeader(BackupFormat::FileType, BackupRecord*, BackupFile*);
   void sendBackupRef(Signal* signal, BackupRecordPtr ptr, Uint32 errorCode);
-  void sendBackupRef(BlockReference ref, Signal *signal,
+  void sendBackupRef(BlockReference ref, Uint32 flags, Signal *signal,
 		     Uint32 senderData, Uint32 errorCode);
   void dumpUsedResources();
   void cleanup(Signal*, BackupRecordPtr ptr);
