@@ -25,7 +25,6 @@
 #include <AttributeHeader.hpp>
 #include <Interpreter.hpp>
 #include <signaldata/FsConf.hpp>
-#include <signaldata/FsRef.hpp>
 #include <signaldata/FsRemoveReq.hpp>
 #include <signaldata/TupCommit.hpp>
 #include <signaldata/TupKey.hpp>
@@ -76,25 +75,8 @@ Dbtup::Dbtup(const class Configuration & conf)
   c_storedProcPool(),
   c_buildIndexList(c_buildIndexPool)
 {
-  Uint32 log_page_size= 0;
   BLOCK_CONSTRUCTOR(Dbtup);
 
-  const ndb_mgm_configuration_iterator * p = conf.getOwnConfigIterator();
-  ndbrequire(p != 0);
-
-  ndb_mgm_get_int_parameter(p, CFG_DB_UNDO_DATA_BUFFER,  
-			    &log_page_size);
-
-  /**
-   * Always set page size in half MBytes
-   */
-  cnoOfUndoPage= (log_page_size / sizeof(UndoPage));
-  Uint32 mega_byte_part= cnoOfUndoPage & 15;
-  if (mega_byte_part != 0) {
-    jam();
-    cnoOfUndoPage+= (16 - mega_byte_part);
-  }
-
   addRecSignal(GSN_DEBUG_SIG, &Dbtup::execDEBUG_SIG);
   addRecSignal(GSN_CONTINUEB, &Dbtup::execCONTINUEB);
 
@@ -117,13 +99,9 @@ Dbtup::Dbtup(const class Configuration & conf)
   addRecSignal(GSN_TUP_SRREQ, &Dbtup::execTUP_SRREQ);
   addRecSignal(GSN_TUP_PREPLCPREQ, &Dbtup::execTUP_PREPLCPREQ);
   addRecSignal(GSN_FSOPENCONF, &Dbtup::execFSOPENCONF);
-  addRecSignal(GSN_FSOPENREF, &Dbtup::execFSOPENREF);
   addRecSignal(GSN_FSCLOSECONF, &Dbtup::execFSCLOSECONF);
-  addRecSignal(GSN_FSCLOSEREF, &Dbtup::execFSCLOSEREF);
   addRecSignal(GSN_FSWRITECONF, &Dbtup::execFSWRITECONF);
-  addRecSignal(GSN_FSWRITEREF, &Dbtup::execFSWRITEREF);
   addRecSignal(GSN_FSREADCONF, &Dbtup::execFSREADCONF);
-  addRecSignal(GSN_FSREADREF, &Dbtup::execFSREADREF);
   addRecSignal(GSN_NDB_STTOR, &Dbtup::execNDB_STTOR);
   addRecSignal(GSN_READ_CONFIG_REQ, &Dbtup::execREAD_CONFIG_REQ, true);
   addRecSignal(GSN_SET_VAR_REQ,  &Dbtup::execSET_VAR_REQ);
@@ -133,7 +111,6 @@ Dbtup::Dbtup(const class Configuration & conf)
   addRecSignal(GSN_DROP_TRIG_REQ,  &Dbtup::execDROP_TRIG_REQ);
 
   addRecSignal(GSN_DROP_TAB_REQ, &Dbtup::execDROP_TAB_REQ);
-  addRecSignal(GSN_FSREMOVEREF, &Dbtup::execFSREMOVEREF);
   addRecSignal(GSN_FSREMOVECONF, &Dbtup::execFSREMOVECONF);
 
   addRecSignal(GSN_TUP_ALLOCREQ, &Dbtup::execTUP_ALLOCREQ);
@@ -143,7 +120,29 @@ Dbtup::Dbtup(const class Configuration & conf)
   // Ordered index related
   addRecSignal(GSN_BUILDINDXREQ, &Dbtup::execBUILDINDXREQ);
 
+  // Tup scan
+  addRecSignal(GSN_ACC_SCANREQ, &Dbtup::execACC_SCANREQ);
+  addRecSignal(GSN_NEXT_SCANREQ, &Dbtup::execNEXT_SCANREQ);
+  addRecSignal(GSN_ACC_CHECK_SCAN, &Dbtup::execACC_CHECK_SCAN);
+
   initData();
+
+  attrbufrec = 0;  
+  checkpointInfo = 0;  
+  diskBufferSegmentInfo = 0;  
+  fragoperrec = 0;
+  fragrecord = 0;  
+  hostBuffer = 0;  
+  localLogInfo = 0;  
+  operationrec = 0;
+  page = 0;  
+  pageRange = 0;
+  pendingFileOpenInfo = 0;  
+  restartInfoRecord = 0;  
+  tablerec = 0;  
+  tableDescriptor = 0;  
+  undoPage = 0;
+  cnoOfPage = cnoOfAllocatedPages = 0;
 }//Dbtup::Dbtup()
 
 Dbtup::~Dbtup() 
@@ -263,12 +262,6 @@ void Dbtup::execFSCLOSECONF(Signal* signal)
   releasePendingFileOpenInfoRecord(pfoPtr);
 }//Dbtup::execFSCLOSECONF()
 
-void Dbtup::execFSCLOSEREF(Signal* signal) 
-{
-  ljamEntry();
-  ndbrequire(false);
-}//Dbtup::execFSCLOSEREF()
-
 void Dbtup::execFSOPENCONF(Signal* signal) 
 {
   PendingFileOpenInfoPtr pfoPtr;
@@ -363,12 +356,6 @@ void Dbtup::execFSOPENCONF(Signal* signal)
   releasePendingFileOpenInfoRecord(pfoPtr);
 }//Dbtup::execFSOPENCONF()
 
-void Dbtup::execFSOPENREF(Signal* signal) 
-{
-  ljamEntry();
-  ndbrequire(false);
-}//Dbtup::execFSOPENREF()
-
 void Dbtup::execFSREADCONF(Signal* signal) 
 {
   DiskBufferSegmentInfoPtr dbsiPtr;
@@ -424,12 +411,6 @@ void Dbtup::execFSREADCONF(Signal* signal)
   }//switch
 }//Dbtup::execFSREADCONF()
 
-void Dbtup::execFSREADREF(Signal* signal) 
-{
-  ljamEntry();
-  ndbrequire(false);
-}//Dbtup::execFSREADREF()
-
 void Dbtup::execFSWRITECONF(Signal* signal) 
 {
   DiskBufferSegmentInfoPtr dbsiPtr;
@@ -478,12 +459,6 @@ void Dbtup::execFSWRITECONF(Signal* signal)
   return;
 }//Dbtup::execFSWRITECONF()
 
-void Dbtup::execFSWRITEREF(Signal* signal) 
-{
-  ljamEntry();
-  ndbrequire(false);
-}//Dbtup::execFSWRITEREF()
-
 void Dbtup::execCONTINUEB(Signal* signal) 
 {
   ljamEntry();
@@ -552,7 +527,7 @@ void Dbtup::execCONTINUEB(Signal* signal)
   case ZREPORT_MEMORY_USAGE:{
     ljam();
     static int c_currentMemUsed = 0;
-    int now = (cnoOfAllocatedPages * 100)/cnoOfPage;
+    int now = cnoOfPage ? (cnoOfAllocatedPages * 100)/cnoOfPage : 0;
     const int thresholds[] = { 100, 90, 80, 0 };
     
     Uint32 i = 0;
@@ -628,6 +603,20 @@ void Dbtup::execREAD_CONFIG_REQ(Signal* signal)
     theConfiguration.getOwnConfigIterator();
   ndbrequire(p != 0);
   
+  Uint32 log_page_size= 0;
+  ndb_mgm_get_int_parameter(p, CFG_DB_UNDO_DATA_BUFFER,  
+			    &log_page_size);
+
+  /**
+   * Always set page size in half MBytes
+   */
+  cnoOfUndoPage= (log_page_size / sizeof(UndoPage));
+  Uint32 mega_byte_part= cnoOfUndoPage & 15;
+  if (mega_byte_part != 0) {
+    jam();
+    cnoOfUndoPage+= (16 - mega_byte_part);
+  }
+
   ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_TUP_FRAG, &cnoOfFragrec));
 
   ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_TUP_OP_RECS, &cnoOfOprec));
@@ -647,12 +636,19 @@ void Dbtup::execREAD_CONFIG_REQ(Signal* signal)
   ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_DB_NO_TRIGGERS, 
 					&noOfTriggers));
 
+  Uint32 nScanOp;       // use TUX config for now
+  ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_TUX_SCAN_OP, &nScanOp));
+
+
   cnoOfTabDescrRec = (cnoOfTabDescrRec & 0xFFFFFFF0) + 16;
+
+  initRecords();
+
   c_storedProcPool.setSize(noOfStoredProc);
   c_buildIndexPool.setSize(c_noOfBuildIndexRec);
   c_triggerPool.setSize(noOfTriggers);
+  c_scanOpPool.setSize(nScanOp);
 
-  initRecords();
   czero = 0;
   cminusOne = czero - 1;
   clastBitMask = 1;
@@ -663,16 +659,33 @@ void Dbtup::execREAD_CONFIG_REQ(Signal* signal)
   initialiseRecordsLab(signal, 0, ref, senderData);
 
   clblPagesPerTick = 50;
-  //ndb_mgm_get_int_parameter(p, CFG_DB_, &clblPagesPerTick);
+  ndb_mgm_get_int_parameter(p, CFG_DB_LCP_DISC_PAGES_TUP_SR, 
+			    &clblPagesPerTick);
 
   clblPagesPerTickAfterSr = 50;
-  //ndb_mgm_get_int_parameter(p, CFG_DB_, &clblPagesPerTickAfterSr);
+  ndb_mgm_get_int_parameter(p, CFG_DB_LCP_DISC_PAGES_TUP, 
+			    &clblPagesPerTickAfterSr);
 
 }//Dbtup::execSIZEALT_REP()
 
 void Dbtup::initRecords() 
 {
+  unsigned i;
+
   // Records with dynamic sizes
+  page = (Page*)allocRecord("Page", 
+			    sizeof(Page), 
+			    cnoOfPage,
+			    false);
+  
+  undoPage = (UndoPage*)allocRecord("UndoPage",
+				    sizeof(UndoPage),
+				    cnoOfUndoPage);
+  
+  operationrec = (Operationrec*)allocRecord("Operationrec",
+					    sizeof(Operationrec),
+					    cnoOfOprec);
+
   attrbufrec = (Attrbufrec*)allocRecord("Attrbufrec", 
 					sizeof(Attrbufrec), 
 					cnoOfAttrbufrec);
@@ -693,6 +706,11 @@ void Dbtup::initRecords()
   fragrecord = (Fragrecord*)allocRecord("Fragrecord",
 					sizeof(Fragrecord), 
 					cnoOfFragrec);
+  
+  for (i = 0; i<cnoOfFragrec; i++) {
+    void * p = &fragrecord[i];
+    new (p) Fragrecord(c_scanOpPool);
+  }
 
   hostBuffer = (HostBuffer*)allocRecord("HostBuffer",
 					sizeof(HostBuffer), 
@@ -702,15 +720,6 @@ void Dbtup::initRecords()
 					    sizeof(LocalLogInfo), 
 					    cnoOfParallellUndoFiles);
 
-  operationrec = (Operationrec*)allocRecord("Operationrec",
-					    sizeof(Operationrec),
-					    cnoOfOprec);
-
-  page = (Page*)allocRecord("Page", 
-			    sizeof(Page), 
-			    cnoOfPage,
-			    false);
-  
   pageRange = (PageRange*)allocRecord("PageRange",
 				      sizeof(PageRange), 
 				      cnoOfPageRangeRec);
@@ -730,7 +739,7 @@ void Dbtup::initRecords()
 				    sizeof(Tablerec), 
 				    cnoOfTablerec);
   
-  for(unsigned i = 0; i<cnoOfTablerec; i++) {
+  for (i = 0; i<cnoOfTablerec; i++) {
     void * p = &tablerec[i];
     new (p) Tablerec(c_triggerPool);
   }
@@ -740,11 +749,6 @@ void Dbtup::initRecords()
 		sizeof(TableDescriptor),
 		cnoOfTabDescrRec);
   
-  undoPage = (UndoPage*)allocRecord("UndoPage",
-				    sizeof(UndoPage),
-				    cnoOfUndoPage);
-  
-  
   // Initialize BAT for interface to file system
   NewVARIABLE* bat = allocateBat(3);
   bat[1].WA = &page->pageWord[0];
