@@ -1145,20 +1145,18 @@ runUniqueNullTransactions(NDBT_Context* ctx, NDBT_Step* step){
   pTrans = pNdb->startTransaction();
   NdbScanOperation * sOp;
   NdbOperation * uOp;
-  NdbResultSet * rs;
   int eof;
   if(!pTrans) goto done;
   sOp = pTrans->getNdbScanOperation(pTab->getName());
   if(!sOp) goto done;
-  rs = sOp->readTuples(NdbScanOperation::LM_Exclusive);
-  if(!rs) goto done;
+  if(sOp->readTuples(NdbScanOperation::LM_Exclusive)) goto done;
   if(pTrans->execute(NoCommit) == -1) goto done;
-  while((eof = rs->nextResult(true)) == 0){
+  while((eof = sOp->nextResult(true)) == 0){
     do {
-      NdbOperation * uOp = rs->updateTuple();
+      NdbOperation * uOp = sOp->updateCurrentTuple();
       if(uOp == 0) goto done;
       uOp->setValue(colId, 0);
-    } while((eof = rs->nextResult(false)) == 0);
+    } while((eof = sOp->nextResult(false)) == 0);
     eof = pTrans->execute(Commit);
     if(eof == -1) goto done;
   }
@@ -1201,6 +1199,48 @@ int runLQHKEYREF(NDBT_Context* ctx, NDBT_Step* step){
   return NDBT_OK;
 }
 
+int 
+runBug21384(NDBT_Context* ctx, NDBT_Step* step)
+{
+  Ndb* pNdb = GETNDB(step);
+  HugoTransactions hugoTrans(*ctx->getTab());
+  NdbRestarter restarter;
+  
+  int loops = ctx->getNumLoops();
+  const int rows = ctx->getNumRecords();
+  const int batchsize = ctx->getProperty("BatchSize", 50);
+  
+  while (loops--)
+  {
+    if(restarter.insertErrorInAllNodes(8037) != 0)
+    {
+      g_err << "Failed to error insert(8037)" << endl;
+      return NDBT_FAILED;
+    }
+    
+    if (hugoTrans.indexReadRecords(pNdb, pkIdxName, rows, batchsize) == 0)
+    {
+      g_err << "Index succeded (it should have failed" << endl;
+      return NDBT_FAILED;
+    }
+    
+    if(restarter.insertErrorInAllNodes(0) != 0)
+    {
+      g_err << "Failed to error insert(0)" << endl;
+      return NDBT_FAILED;
+    }
+    
+    if (hugoTrans.indexReadRecords(pNdb, pkIdxName, rows, batchsize) != 0){
+      g_err << "Index read failed" << endl;
+      return NDBT_FAILED;
+    }
+  }
+  
+  return NDBT_OK;
+}
+
+
+
 NDBT_TESTSUITE(testIndex);
 TESTCASE("CreateAll", 
 	 "Test that we can create all various indexes on each table\n"
@@ -1279,7 +1319,7 @@ TESTCASE("CreateLoadDrop_O",
 TESTCASE("NFNR1", 
 	 "Test that indexes are correctly maintained during node fail and node restart"){ 
   TC_PROPERTY("LoggedIndexes", (unsigned)0);
-  //TC_PROPERTY("Threads", 2);
+  TC_PROPERTY("PauseThreads", 2);
   INITIALIZER(runClearTable);
   INITIALIZER(createRandomIndex);
   INITIALIZER(runLoadTable);
@@ -1294,6 +1334,7 @@ TESTCASE("NFNR1_O",
 	 "Test that indexes are correctly maintained during node fail and node restart"){ 
   TC_PROPERTY("OrderedIndex", 1);
   TC_PROPERTY("LoggedIndexes", (unsigned)0);
+  TC_PROPERTY("PauseThreads", 2);
   INITIALIZER(runClearTable);
   INITIALIZER(createRandomIndex);
   INITIALIZER(runLoadTable);
@@ -1307,6 +1348,7 @@ TESTCASE("NFNR1_O",
 TESTCASE("NFNR2", 
 	 "Test that indexes are correctly maintained during node fail and node restart"){ 
   TC_PROPERTY("LoggedIndexes", (unsigned)0);
+  TC_PROPERTY("PauseThreads", 2);
   INITIALIZER(runClearTable);
   INITIALIZER(createRandomIndex);
   INITIALIZER(createPkIndex);
@@ -1323,13 +1365,14 @@ TESTCASE("NFNR2_O",
 	 "Test that indexes are correctly maintained during node fail and node restart"){ 
   TC_PROPERTY("OrderedIndex", 1);
   TC_PROPERTY("LoggedIndexes", (unsigned)0);
+  TC_PROPERTY("PauseThreads", 1);
   INITIALIZER(runClearTable);
   INITIALIZER(createRandomIndex);
   INITIALIZER(createPkIndex);
   INITIALIZER(runLoadTable);
   STEP(runRestarts);
   STEP(runTransactions2);
-  STEP(runTransactions2);
+  //STEP(runTransactions2);
   FINALIZER(runVerifyIndex);
   FINALIZER(createRandomIndex_Drop);
   FINALIZER(createPkIndex_Drop);
@@ -1338,6 +1381,7 @@ TESTCASE("NFNR2_O",
 TESTCASE("NFNR3", 
 	 "Test that indexes are correctly maintained during node fail and node restart"){ 
   TC_PROPERTY("LoggedIndexes", (unsigned)0);
+  TC_PROPERTY("PauseThreads", 2);
   INITIALIZER(runClearTable);
   INITIALIZER(createRandomIndex);
   INITIALIZER(createPkIndex);
@@ -1353,6 +1397,7 @@ TESTCASE("NFNR3_O",
 	 "Test that indexes are correctly maintained during node fail and node restart"){ 
   TC_PROPERTY("OrderedIndex", 1);
   TC_PROPERTY("LoggedIndexes", (unsigned)0);
+  TC_PROPERTY("PauseThreads", 2);
   INITIALIZER(runClearTable);
   INITIALIZER(createRandomIndex);
   INITIALIZER(createPkIndex);
@@ -1367,6 +1412,7 @@ TESTCASE("NFNR3_O",
 TESTCASE("NFNR4", 
 	 "Test that indexes are correctly maintained during node fail and node restart"){ 
   TC_PROPERTY("LoggedIndexes", (unsigned)0);
+  TC_PROPERTY("PauseThreads", 4);
   INITIALIZER(runClearTable);
   INITIALIZER(createRandomIndex);
   INITIALIZER(createPkIndex);
@@ -1385,6 +1431,7 @@ TESTCASE("NFNR4_O",
 	 "Test that indexes are correctly maintained during node fail and node restart"){ 
   TC_PROPERTY("OrderedIndex", 1);
   TC_PROPERTY("LoggedIndexes", (unsigned)0);
+  TC_PROPERTY("PauseThreads", 4);
   INITIALIZER(runClearTable);
   INITIALIZER(createRandomIndex);
   INITIALIZER(createPkIndex);
@@ -1507,6 +1554,16 @@ TESTCASE("UniqueNull",
   FINALIZER(createPkIndex_Drop);
   FINALIZER(runClearTable);
 }
+TESTCASE("Bug21384", 
+	 "Test that unique indexes and nulls"){ 
+  TC_PROPERTY("LoggedIndexes", (unsigned)0);
+  INITIALIZER(runClearTable);
+  INITIALIZER(createPkIndex);
+  INITIALIZER(runLoadTable);
+  STEP(runBug21384);
+  FINALIZER(createPkIndex_Drop);
+  FINALIZER(runClearTable);
+}
 NDBT_TESTSUITE_END(testIndex);
 
 int main(int argc, const char** argv){
