@@ -32,16 +32,22 @@ public:
   Item_str_func(Item *a,Item *b,Item *c,Item *d, Item* e) :Item_func(a,b,c,d,e) {decimals=NOT_FIXED_DEC; }
   Item_str_func(List<Item> &list) :Item_func(list) {decimals=NOT_FIXED_DEC; }
   longlong val_int();
-  double val();
+  double val_real();
+  my_decimal *val_decimal(my_decimal *);
   enum Item_result result_type () const { return STRING_RESULT; }
   void left_right_max_length();
+  String *check_well_formed_result(String *str);
+  bool fix_fields(THD *thd, Item **ref);
 };
 
 class Item_func_md5 :public Item_str_func
 {
   String tmp_value;
 public:
-  Item_func_md5(Item *a) :Item_str_func(a) {}
+  Item_func_md5(Item *a) :Item_str_func(a)
+  {
+    collation.set(&my_charset_bin);
+  }
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "md5"; }
@@ -51,7 +57,10 @@ public:
 class Item_func_sha :public Item_str_func
 {
 public:
-  Item_func_sha(Item *a) :Item_str_func(a) {}  
+  Item_func_sha(Item *a) :Item_str_func(a)
+  {
+    collation.set(&my_charset_bin);
+  }
   String *val_str(String *);    
   void fix_length_and_dec();      
   const char *func_name() const { return "sha"; }	
@@ -100,6 +109,7 @@ public:
 
 class Item_func_reverse :public Item_str_func
 {
+  String tmp_value;
 public:
   Item_func_reverse(Item *a) :Item_str_func(a) {}
   String *val_str(String *);
@@ -134,13 +144,14 @@ public:
 
 class Item_str_conv :public Item_str_func
 {
+protected:
+  uint multiply;
+  uint (*converter)(CHARSET_INFO *cs, char *src, uint srclen,
+                                      char *dst, uint dstlen);
+  String tmp_value;
 public:
   Item_str_conv(Item *item) :Item_str_func(item) {}
-  void fix_length_and_dec()
-  {
-    collation.set(args[0]->collation);
-    max_length = args[0]->max_length;
-  }
+  String *val_str(String *);
 };
 
 
@@ -148,16 +159,28 @@ class Item_func_lcase :public Item_str_conv
 {
 public:
   Item_func_lcase(Item *item) :Item_str_conv(item) {}
-  String *val_str(String *);
   const char *func_name() const { return "lcase"; }
+  void fix_length_and_dec()
+  {
+    collation.set(args[0]->collation);
+    multiply= collation.collation->casedn_multiply;
+    converter= collation.collation->cset->casedn;
+    max_length= args[0]->max_length * multiply;
+  }
 };
 
 class Item_func_ucase :public Item_str_conv
 {
 public:
   Item_func_ucase(Item *item) :Item_str_conv(item) {}
-  String *val_str(String *);
   const char *func_name() const { return "ucase"; }
+  void fix_length_and_dec()
+  {
+    collation.set(args[0]->collation);
+    multiply= collation.collation->caseup_multiply;
+    converter= collation.collation->cset->caseup;
+    max_length= args[0]->max_length * multiply;
+  }
 };
 
 
@@ -202,7 +225,7 @@ public:
   Item_func_substr_index(Item *a,Item *b,Item *c) :Item_str_func(a,b,c) {}
   String *val_str(String *);
   void fix_length_and_dec();
-  const char *func_name() const { return "substr_index"; }
+  const char *func_name() const { return "substring_index"; }
 };
 
 
@@ -217,6 +240,8 @@ public:
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "trim"; }
+  void print(String *str);
+  virtual const char *mode_name() const { return "both"; }
 };
 
 
@@ -227,6 +252,7 @@ public:
   Item_func_ltrim(Item *a) :Item_func_trim(a) {}
   String *val_str(String *);
   const char *func_name() const { return "ltrim"; }
+  const char *mode_name() const { return "leading"; }
 };
 
 
@@ -237,6 +263,7 @@ public:
   Item_func_rtrim(Item *a) :Item_func_trim(a) {}
   String *val_str(String *);
   const char *func_name() const { return "rtrim"; }
+  const char *mode_name() const { return "trailing"; }
 };
 
 
@@ -245,7 +272,7 @@ public:
   Returns strcat('*', octet2hex(sha1(sha1(password)))). '*' stands for new
   password format, sha1(sha1(password) is so-called hash_stage2 value.
   Length of returned string is always 41 byte. To find out how entire
-  authentification procedure works, see comments in password.c.
+  authentication procedure works, see comments in password.c.
 */
 
 class Item_func_password :public Item_str_func
@@ -305,12 +332,24 @@ public:
 class Item_func_encrypt :public Item_str_func
 {
   String tmp_value;
+
+  /* Encapsulate common constructor actions */
+  void constructor_helper()
+  {
+    collation.set(&my_charset_bin);
+  }
 public:
-  Item_func_encrypt(Item *a) :Item_str_func(a) {}
-  Item_func_encrypt(Item *a, Item *b): Item_str_func(a,b) {}
+  Item_func_encrypt(Item *a) :Item_str_func(a)
+  {
+    constructor_helper();
+  }
+  Item_func_encrypt(Item *a, Item *b): Item_str_func(a,b)
+  {
+    constructor_helper();
+  }
   String *val_str(String *);
   void fix_length_and_dec() { maybe_null=1; max_length = 13; }
-  const char *func_name() const { return "ecrypt"; }
+  const char *func_name() const { return "encrypt"; }
 };
 
 #include "sql_crypt.h"
@@ -344,8 +383,15 @@ public:
   Item_func_sysconst()
   { collation.set(system_charset_info,DERIVATION_SYSCONST); }
   Item *safe_charset_converter(CHARSET_INFO *tocs);
+  /*
+    Used to create correct Item name in new converted item in
+    safe_charset_converter, return string representation of this function
+    call
+  */
+  virtual const char *fully_qualified_func_name() const = 0;
 };
 
+
 class Item_func_database :public Item_func_sysconst
 {
 public:
@@ -357,18 +403,46 @@ public:
     maybe_null=1;
   }
   const char *func_name() const { return "database"; }
+  const char *fully_qualified_func_name() const { return "database()"; }
 };
 
+
 class Item_func_user :public Item_func_sysconst
 {
+protected:
+  bool init (const char *user, const char *host);
+
 public:
-  Item_func_user() :Item_func_sysconst() {}
-  String *val_str(String *);
-  void fix_length_and_dec() 
-  { 
-    max_length= (USERNAME_LENGTH+HOSTNAME_LENGTH+1)*system_charset_info->mbmaxlen;
+  Item_func_user()
+  {
+    str_value.set("", 0, system_charset_info);
+  }
+  String *val_str(String *)
+  {
+    DBUG_ASSERT(fixed == 1);
+    return (null_value ? 0 : &str_value);
+  }
+  bool fix_fields(THD *thd, Item **ref);
+  void fix_length_and_dec()
+  {
+    max_length= ((USERNAME_LENGTH + HOSTNAME_LENGTH + 1) *
+                 system_charset_info->mbmaxlen);
   }
   const char *func_name() const { return "user"; }
+  const char *fully_qualified_func_name() const { return "user()"; }
+};
+
+
+class Item_func_current_user :public Item_func_user
+{
+  Name_resolution_context *context;
+
+public:
+  Item_func_current_user(Name_resolution_context *context_arg)
+    : context(context_arg) {}
+  bool fix_fields(THD *thd, Item **ref);
+  const char *func_name() const { return "current_user"; }
+  const char *fully_qualified_func_name() const { return "current_user()"; }
 };
 
 
@@ -387,7 +461,7 @@ class Item_func_elt :public Item_str_func
 {
 public:
   Item_func_elt(List<Item> &list) :Item_str_func(list) {}
-  double val();
+  double val_real();
   longlong val_int();
   String *val_str(String *str);
   void fix_length_and_dec();
@@ -403,13 +477,12 @@ class Item_func_make_set :public Item_str_func
 public:
   Item_func_make_set(Item *a,List<Item> &list) :Item_str_func(list),item(a) {}
   String *val_str(String *str);
-  bool fix_fields(THD *thd, TABLE_LIST *tlist, Item **ref)
+  bool fix_fields(THD *thd, Item **ref)
   {
     DBUG_ASSERT(fixed == 0);
-    return (!item->fixed &&
-            item->fix_fields(thd, tlist, &item) ||
+    return ((!item->fixed && item->fix_fields(thd, &item)) ||
 	    item->check_cols(1) ||
-	    Item_func::fix_fields(thd, tlist, ref));
+	    Item_func::fix_fields(thd, ref));
   }
   void split_sum_func(THD *thd, Item **ref_pointer_array, List<Item> &fields);
   void fix_length_and_dec();
@@ -421,6 +494,7 @@ public:
     return item->walk(processor, arg) ||
       Item_str_func::walk(processor, arg);
   }
+  Item *transform(Item_transformer transformer, byte *arg);
   void print(String *str);
 };
 
@@ -434,7 +508,9 @@ public:
   void fix_length_and_dec()
   {
     collation.set(default_charset());
-    max_length=args[0]->max_length+(args[0]->max_length-args[0]->decimals)/3;
+    uint char_length= args[0]->max_length/args[0]->collation.collation->mbmaxlen;
+    max_length= ((char_length + (char_length-args[0]->decimals)/3) *
+                 collation.collation->mbmaxlen);
   }
   const char *func_name() const { return "format"; }
   void print(String *);
@@ -444,12 +520,14 @@ public:
 class Item_func_char :public Item_str_func
 {
 public:
-  Item_func_char(List<Item> &list) :Item_str_func(list) {}
+  Item_func_char(List<Item> &list) :Item_str_func(list)
+  { collation.set(&my_charset_bin); }
+  Item_func_char(List<Item> &list, CHARSET_INFO *cs) :Item_str_func(list)
+  { collation.set(cs); }  
   String *val_str(String *);
   void fix_length_and_dec() 
-  { 
-    collation.set(default_charset());
-    maybe_null=0; max_length=arg_count;
+  {
+    max_length= arg_count * collation.collation->mbmaxlen;
   }
   const char *func_name() const { return "char"; }
 };
@@ -499,7 +577,7 @@ public:
   void fix_length_and_dec()
   {
     collation.set(default_charset());
-    decimals=0; max_length=64;
+    max_length= 64;
   }
 };
 
@@ -554,6 +632,7 @@ public:
     max_length=args[0]->max_length;
   }
   void print(String *str);
+  const char *func_name() const { return "cast_as_binary"; }
 };
 
 
@@ -611,10 +690,41 @@ public:
 
 class Item_func_conv_charset :public Item_str_func
 {
-  CHARSET_INFO *conv_charset;
+  bool use_cached_value;
 public:
-  Item_func_conv_charset(Item *a, CHARSET_INFO *cs) :Item_str_func(a)
-  { conv_charset=cs; }
+  bool safe;
+  CHARSET_INFO *conv_charset; // keep it public
+  Item_func_conv_charset(Item *a, CHARSET_INFO *cs) :Item_str_func(a) 
+  { conv_charset= cs; use_cached_value= 0; safe= 0; }
+  Item_func_conv_charset(Item *a, CHARSET_INFO *cs, bool cache_if_const) 
+    :Item_str_func(a) 
+  {
+    DBUG_ASSERT(args[0]->fixed);
+    conv_charset= cs;
+    if (cache_if_const && args[0]->const_item())
+    {
+      uint errors= 0;
+      String tmp, *str= args[0]->val_str(&tmp);
+      if (!str || str_value.copy(str->ptr(), str->length(),
+                                 str->charset(), conv_charset, &errors))
+        null_value= 1;
+      use_cached_value= 1;
+      str_value.mark_as_const();
+      safe= (errors == 0);
+    }
+    else
+    {
+      use_cached_value= 0;
+      /*
+        Conversion from and to "binary" is safe.
+        Conversion to Unicode is safe.
+        Other kind of conversions are potentially lossy.
+      */
+      safe= (args[0]->collation.collation == &my_charset_bin ||
+             cs == &my_charset_bin ||
+             (cs->state & MY_CS_UNICODE));
+    }
+  }
   String *val_str(String *);
   void fix_length_and_dec();
   const char *func_name() const { return "convert"; }
@@ -629,7 +739,13 @@ public:
   void fix_length_and_dec();
   bool eq(const Item *item, bool binary_cmp) const;
   const char *func_name() const { return "collate"; }
-  void print(String *str) { print_op(str); }
+  enum Functype functype() const { return COLLATE_FUNC; }
+  void print(String *str);
+  Item_field *filed_for_view_update()
+  {
+    /* this function is transparent for view updating */
+    return args[0]->filed_for_view_update();
+  }
 };
 
 class Item_func_charset :public Item_str_func
@@ -703,7 +819,7 @@ class Item_func_uncompress: public Item_str_func
   String buffer;
 public:
   Item_func_uncompress(Item *a): Item_str_func(a){}
-  void fix_length_and_dec(){max_length= MAX_BLOB_WIDTH;}
+  void fix_length_and_dec(){ maybe_null= 1; max_length= MAX_BLOB_WIDTH; }
   const char *func_name() const{return "uncompress";}
   String *val_str(String *) ZLIB_DEPENDED_FUNCTION
 };
@@ -715,7 +831,12 @@ public:
   Item_func_uuid(): Item_str_func() {}
   void fix_length_and_dec() {
     collation.set(system_charset_info);
-    max_length= UUID_LENGTH;
+    /*
+       NOTE! uuid() should be changed to use 'ascii'
+       charset when hex(), format(), md5(), etc, and implicit
+       number-to-string conversion will use 'ascii'
+    */
+    max_length= UUID_LENGTH * system_charset_info->mbmaxlen;
   }
   const char *func_name() const{ return "uuid"; }
   String *val_str(String *);
