@@ -30,6 +30,6 @@
 
 char *llstr(longlong value,char *buff)
 {
-  longlong2str(value,buff,-10);
+  longlong10_to_str(value,buff,-10);
   return buff;
 }
